import copy
import email
import fnmatch
import json
import logging
import os
from abc import ABC, abstractmethod
from enum import Enum
from pathlib import Path
from shutil import copy2
from threading import Event
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union

import yaml
from robot.api.deco import keyword, library
from robot.libraries.BuiltIn import BuiltIn

from RPA.core.helpers import import_by_name, required_env
from RPA.core.logger import deprecation
from RPA.core.notebook import notebook_print
from RPA.FileSystem import FileSystem
from RPA.Robocorp.utils import (
    JSONType,
    Requests,
    get_dot_value,
    is_json_equal,
    json_dumps,
    resolve_path,
    set_dot_value,
    truncate,
    url_join,
)


UNDEFINED = object()  # Undefined default value
ENCODING = "utf-8"

_STRINGS_TYPE = Union[str, Tuple[str, ...]]
AUTO_PARSE_EMAIL_TYPE = Optional[Dict[_STRINGS_TYPE, _STRINGS_TYPE]]
AUTO_PARSE_EMAIL_DEFAULT = {
    # Source payload keys or file names -> destination payload keys with the parsed
    # e-mail content.
    ("email.text", "__mail.html"): ("email.body", "parsedEmail.Body"),
    "rawEmail": "parsedEmail",
}


class State(Enum):
    """Work item state. (set when released)"""

    DONE = "COMPLETED"
    FAILED = "FAILED"


class Error(Enum):
    """Failed work item error type."""

    BUSINESS = "BUSINESS"  # wrong/missing data, shouldn't be retried
    APPLICATION = "APPLICATION"  # logic issue/timeout, can be retried


class EmptyQueue(IndexError):
    """Raised when trying to load an input item and none available."""


class BaseAdapter(ABC):
    """Abstract base class for work item adapters."""

    @abstractmethod
    def reserve_input(self) -> str:
        """Get next work item ID from the input queue and reserve it."""
        raise NotImplementedError

    @abstractmethod
    def release_input(
        self, item_id: str, state: State, exception: Optional[dict] = None
    ):
        """Release the lastly retrieved input work item and set state."""
        raise NotImplementedError

    @abstractmethod
    def create_output(self, parent_id: str, payload: Optional[JSONType] = None) -> str:
        """Create new output for work item, and return created ID."""
        raise NotImplementedError

    @abstractmethod
    def load_payload(self, item_id: str) -> JSONType:
        """Load JSON payload from work item."""
        raise NotImplementedError

    @abstractmethod
    def save_payload(self, item_id: str, payload: JSONType):
        """Save JSON payload to work item."""
        raise NotImplementedError

    @abstractmethod
    def list_files(self, item_id: str) -> List[str]:
        """List attached files in work item."""
        raise NotImplementedError

    @abstractmethod
    def get_file(self, item_id: str, name: str) -> bytes:
        """Read file's contents from work item."""
        raise NotImplementedError

    @abstractmethod
    def add_file(self, item_id: str, name: str, *, original_name: str, content: bytes):
        """Attach file to work item."""
        raise NotImplementedError

    @abstractmethod
    def remove_file(self, item_id: str, name: str):
        """Remove attached file from work item."""
        raise NotImplementedError


class RobocorpAdapter(BaseAdapter):
    """Adapter for saving/loading work items from Robocorp Control Room.

    Required environment variables:

    * RC_API_WORKITEM_HOST:     Work item API hostname
    * RC_API_WORKITEM_TOKEN:    Work item API access token

    * RC_API_PROCESS_HOST:      Process API hostname
    * RC_API_PROCESS_TOKEN:     Process API access token

    * RC_WORKSPACE_ID:          Control room workspace ID
    * RC_PROCESS_ID:            Control room process ID
    * RC_PROCESS_RUN_ID:        Control room process run ID
    * RC_ROBOT_RUN_ID:          Control room robot run ID

    * RC_WORKITEM_ID:           Control room work item ID (input)
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # IDs identifying the current robot run and its input.
        self._workspace_id = required_env("RC_WORKSPACE_ID")
        self._process_run_id = required_env("RC_PROCESS_RUN_ID")
        self._step_run_id = required_env("RC_ACTIVITY_RUN_ID")
        self._initial_item_id: Optional[str] = required_env("RC_WORKITEM_ID")

        self._workitem_requests = self._process_requests = None
        self._init_workitem_requests()
        self._init_process_requests()

    def _init_workitem_requests(self):
        # Endpoint for old work items API.
        workitem_host = required_env("RC_API_WORKITEM_HOST")
        workitem_token = required_env("RC_API_WORKITEM_TOKEN")
        route_prefix = (
            url_join(
                workitem_host, "json-v1", "workspaces", self._workspace_id, "workitems"
            )
            + "/"
        )
        default_headers = {
            "Authorization": f"Bearer {workitem_token}",
            "Content-Type": "application/json",
        }
        logging.info("Work item API route prefix: %s", route_prefix)
        self._workitem_requests = Requests(
            route_prefix, default_headers=default_headers
        )

    def _init_process_requests(self):
        # Endpoint for the new process API.
        process_host = required_env("RC_API_PROCESS_HOST")
        process_token = required_env("RC_API_PROCESS_TOKEN")
        process_id = required_env("RC_PROCESS_ID")
        route_prefix = (
            url_join(
                process_host,
                "process-v1",
                "workspaces",
                self._workspace_id,
                "processes",
                process_id,
            )
            + "/"
        )
        default_headers = {
            "Authorization": f"Bearer {process_token}",
            "Content-Type": "application/json",
        }
        logging.info("Process API route prefix: %s", route_prefix)
        self._process_requests = Requests(route_prefix, default_headers=default_headers)

    def _pop_item(self):
        # Get the next input work item from the cloud queue.
        url = url_join(
            "runs",
            self._process_run_id,
            "robotRuns",
            self._step_run_id,
            "reserve-next-work-item",
        )
        logging.info("Reserving new input work item from: %s", url)
        response = self._process_requests.post(url)
        return response.json()["workItemId"]

    def reserve_input(self) -> str:
        if self._initial_item_id:
            item_id = self._initial_item_id
            self._initial_item_id = None
            return item_id

        item_id = self._pop_item()
        if not item_id:
            raise EmptyQueue("No work items in the input queue")
        return item_id

    def release_input(
        self, item_id: str, state: State, exception: Optional[dict] = None
    ):
        # Release the current input work item in the cloud queue.
        url = url_join(
            "runs",
            self._process_run_id,
            "robotRuns",
            self._step_run_id,
            "release-work-item",
        )
        body = {"workItemId": item_id, "state": state.value}
        if exception:
            for key, value in list(exception.items()):
                # All values are (and should be) strings (if not `None`).
                value = value.strip() if value else value
                if value:
                    exception[key] = value  # keep the stripped string value
                else:
                    # Exclude `None` & empty string values.
                    del exception[key]
            body["exception"] = exception

        log_func = logging.error if state == State.FAILED else logging.info
        log_func(
            "Releasing %s input work item %r into %r with exception: %s",
            state.value,
            item_id,
            url,
            exception,
        )
        self._process_requests.post(url, json=body)

    def create_output(self, parent_id: str, payload: Optional[JSONType] = None) -> str:
        # Putting "output" for the current input work item identified by `parent_id`.
        url = url_join("work-items", parent_id, "output")
        body = {"payload": payload}

        logging.info("Creating output item: %s", url)
        response = self._process_requests.post(url, json=body)
        return response.json()["id"]

    def load_payload(self, item_id: str) -> JSONType:
        url = url_join(item_id, "data")

        def handle_error(resp):
            # NOTE: The API might return 404 if no payload is attached to the work
            # item.
            if not (resp.ok or resp.status_code == 404):
                self._workitem_requests.handle_error(resp)

        logging.info("Loading work item payload from: %s", url)
        response = self._workitem_requests.get(url, _handle_error=handle_error)
        return response.json() if response.ok else {}

    def save_payload(self, item_id: str, payload: JSONType):
        url = url_join(item_id, "data")

        logging.info("Saving work item payload to: %s", url)
        self._workitem_requests.put(url, json=payload)

    def list_files(self, item_id: str) -> List[str]:
        url = url_join(item_id, "files")

        logging.info("Listing work item files at: %s", url)
        response = self._workitem_requests.get(url)

        return [item["fileName"] for item in response.json()]

    def get_file(self, item_id: str, name: str) -> bytes:
        # Robocorp API returns URL for S3 download.
        file_id = self.file_id(item_id, name)
        url = url_join(item_id, "files", file_id)

        logging.info("Downloading work item file at: %s", url)
        response = self._workitem_requests.get(url)
        file_url = response.json()["url"]

        # Perform the actual file download.
        response = self._workitem_requests.get(
            file_url,
            _handle_error=lambda resp: resp.raise_for_status(),
            _sensitive=True,
            headers={},
        )
        return response.content

    def add_file(self, item_id: str, name: str, *, original_name: str, content: bytes):
        # Note that here the `original_name` is useless here. (used with `FileAdapter`
        #   only)
        del original_name

        # Robocorp API returns pre-signed POST details for S3 upload.
        url = url_join(item_id, "files")
        body = {"fileName": str(name), "fileSize": len(content)}
        logging.info(
            "Adding work item file into: %s (name: %s, size: %d)",
            url,
            body["fileName"],
            body["fileSize"],
        )
        response = self._workitem_requests.post(url, json=body)
        data = response.json()

        # Perform the actual file upload.
        url = data["url"]
        fields = data["fields"]
        files = {"file": (name, content)}
        self._workitem_requests.post(
            url,
            _handle_error=lambda resp: resp.raise_for_status(),
            _sensitive=True,
            headers={},
            data=fields,
            files=files,
        )

    def remove_file(self, item_id: str, name: str):
        file_id = self.file_id(item_id, name)
        url = url_join(item_id, "files", file_id)
        self._workitem_requests.delete(url)

    def file_id(self, item_id: str, name: str) -> str:
        url = url_join(item_id, "files")
        response = self._workitem_requests.get(url)

        files = response.json()
        if not files:
            raise FileNotFoundError("No files in work item")

        matches = [item for item in files if item["fileName"] == name]
        if not matches:
            raise FileNotFoundError(
                "File with name '{name}' not in: {names}".format(
                    name=name, names=", ".join(item["fileName"] for item in files)
                )
            )

        # Duplicate filenames should never exist,
        # but use last item just in case
        return matches[-1]["fileId"]


class FileAdapter(BaseAdapter):
    """Adapter for simulating work item input queues.

    Reads inputs from the given database file, and writes
    all created output items into an adjacent file
    with the suffix ``<filename>.output.json``. If the output path is provided by an
    env var explicitly, then the file will be saved with the provided path and name.

    Reads and writes all work item files from/to the same parent
    folder as the given input database.

    Optional environment variables:

    * RPA_INPUT_WORKITEM_PATH:  Path to work items input database file
    * RPA_OUTPUT_WORKITEM_PATH:  Path to work items output database file
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._input_path = UNDEFINED
        self._output_path = UNDEFINED

        self.inputs: List[Dict[str, Any]] = self.load_database()
        self.outputs: List[Dict[str, Any]] = []
        self.index: int = 0

    def _get_item(self, item_id: str) -> Tuple[str, Dict[str, Any]]:
        # The work item ID is analogue to inputs/outputs list queues index.
        idx = int(item_id)
        if idx < len(self.inputs):
            return "input", self.inputs[idx]

        if idx < (len(self.inputs) + len(self.outputs)):
            return "output", self.outputs[idx - len(self.inputs)]

        raise ValueError(f"Unknown work item ID: {item_id}")

    def reserve_input(self) -> str:
        if self.index >= len(self.inputs):
            raise EmptyQueue("No work items in the input queue")

        try:
            return str(self.index)
        finally:
            self.index += 1

    def release_input(
        self, item_id: str, state: State, exception: Optional[dict] = None
    ):
        # Nothing happens for now on releasing local dev input Work Items.
        log_func = logging.error if state == State.FAILED else logging.info
        log_func(
            "Releasing item %r with %s state and exception: %s",
            item_id,
            state.value,
            exception,
        )

    @property
    def input_path(self) -> Optional[Path]:
        if self._input_path is UNDEFINED:
            # pylint: disable=invalid-envvar-default
            old_path = os.getenv("RPA_WORKITEMS_PATH")
            if old_path:
                deprecation(
                    "Work items load - Old path style usage detected, please use the "
                    "'RPA_INPUT_WORKITEM_PATH' env var instead "
                    "(more details under documentation: https://robocorp.com/docs/development-guide/control-room/data-pipeline#developing-with-work-items-locally)"  # noqa: E501
                )
            path = os.getenv("RPA_INPUT_WORKITEM_PATH", default=old_path)
            if path:
                logging.info("Resolving input path: %s", path)
                self._input_path = resolve_path(path)
            else:
                # Will raise `TypeError` during inputs loading and will populate the
                # list with one empty initial input.
                self._input_path = None

        return self._input_path

    @property
    def output_path(self) -> Path:
        if self._output_path is UNDEFINED:
            # This is usually set once per loaded input work item.
            new_path = os.getenv("RPA_OUTPUT_WORKITEM_PATH")
            if new_path:
                self._output_path = resolve_path(new_path)
                self._output_path.parent.mkdir(parents=True, exist_ok=True)
            else:
                deprecation(
                    "Work items save - Old path style usage detected, please use the "
                    "'RPA_OUTPUT_WORKITEM_PATH' env var instead "
                    "(more details under documentation: https://robocorp.com/docs/development-guide/control-room/data-pipeline#developing-with-work-items-locally)"  # noqa: E501
                )
                if not self.input_path:
                    raise RuntimeError(
                        "You must provide a path for at least one of the input or "
                        "output work items files"
                    )
                self._output_path = self.input_path.with_suffix(".output.json")

        return self._output_path

    def _save_to_disk(self, source: str) -> None:
        if source == "input":
            if not self.input_path:
                raise RuntimeError(
                    "Can't save an input item without a path defined, use "
                    "'RPA_INPUT_WORKITEM_PATH' env for this matter"
                )
            path = self.input_path
            data = self.inputs
        else:
            path = self.output_path
            data = self.outputs

        with open(path, "w", encoding=ENCODING) as fd:
            fd.write(json_dumps(data, indent=4))

        logging.info("Saved into %s file: %s", source, path)

    def create_output(self, _: str, payload: Optional[JSONType] = None) -> str:
        # Note that the `parent_id` is not used during local development.
        item: Dict[str, Any] = {"payload": payload, "files": {}}
        self.outputs.append(item)

        self._save_to_disk("output")
        return str(len(self.inputs) + len(self.outputs) - 1)  # new output work item ID

    def load_payload(self, item_id: str) -> JSONType:
        _, item = self._get_item(item_id)
        return item.get("payload", {})

    def save_payload(self, item_id: str, payload: JSONType):
        source, item = self._get_item(item_id)
        item["payload"] = payload
        self._save_to_disk(source)

    def list_files(self, item_id: str) -> List[str]:
        _, item = self._get_item(item_id)
        files = item.get("files", {})
        return list(files.keys())

    def get_file(self, item_id: str, name: str) -> bytes:
        source, item = self._get_item(item_id)
        files = item.get("files", {})

        path = files[name]
        if not Path(path).is_absolute():
            parent = (
                self.input_path.parent if source == "input" else self.output_path.parent
            )
            path = parent / path

        with open(path, "rb") as infile:
            return infile.read()

    def add_file(self, item_id: str, name: str, *, original_name: str, content: bytes):
        source, item = self._get_item(item_id)
        files = item.setdefault("files", {})

        parent = (
            self.input_path.parent if source == "input" else self.output_path.parent
        )
        path = parent / original_name  # the file on disk will keep its original name
        with open(path, "wb") as fd:
            fd.write(content)
        logging.info("Created file: %s", path)
        files[name] = original_name  # file path relative to the work item

        self._save_to_disk(source)

    def remove_file(self, item_id: str, name: str):
        source, item = self._get_item(item_id)
        files = item.get("files", {})

        path = files[name]
        logging.info("Would remove file: %s", path)
        # Note that the file doesn't get removed from disk as well.
        del files[name]

        self._save_to_disk(source)

    def load_database(self) -> List:
        try:
            try:
                with open(self.input_path, "r", encoding=ENCODING) as infile:
                    data = json.load(infile)
            except (TypeError, FileNotFoundError):
                logging.warning("No input work items file found: %s", self.input_path)
                data = []

            if isinstance(data, list):
                assert all(
                    isinstance(d, dict) for d in data
                ), "Items should be dictionaries"
                if len(data) == 0:
                    data.append({"payload": {}})
                return data

            # Attempt to migrate from old format
            assert isinstance(data, dict), "Not a list or dictionary"
            deprecation("Work items file as mapping is deprecated")
            workspace = next(iter(data.values()))
            work_item = next(iter(workspace.values()))
            return [{"payload": work_item}]
        except Exception as exc:  # pylint: disable=broad-except
            logging.exception("Invalid work items file because of: %s", exc)
            return [{"payload": {}}]


class WorkItem:
    """Base class for input and output work items.

    :param adapter:   Adapter instance
    :param item_id:   Work item ID (optional)
    :param parent_id: Parent work item's ID (optional)
    """

    def __init__(self, adapter, item_id=None, parent_id=None):
        #: Adapter for loading/saving content
        self.adapter = adapter
        #: This item's and/or parent's ID
        self.id: Optional[str] = item_id
        self.parent_id: Optional[str] = parent_id
        assert self.id is not None or self.parent_id is not None
        #: Item's state on release; can be set once
        self.state: Optional[State] = None
        #: Remote JSON payload, and queued changes
        self._payload: JSONType = {}
        self._payload_cache: JSONType = {}
        #: Remote attached files, and queued changes
        self._files: List[str] = []
        self._files_to_add: Dict[str, Path] = {}
        self._files_to_remove: List[str] = []

    def __repr__(self):
        payload = truncate(str(self.payload), 64)
        files = len(self.files)
        return f"WorkItem(id={self.id}, payload={payload}, files={files})"

    @property
    def is_dirty(self):
        """Check if work item has unsaved changes."""
        return (
            self.id is None
            or not is_json_equal(self._payload, self._payload_cache)
            or self._files_to_add
            or self._files_to_remove
        )

    @property
    def payload(self):
        return self._payload_cache

    @payload.setter
    def payload(self, value):
        self._payload_cache = value

    @property
    def files(self):
        """List of filenames, including local files pending upload and
        excluding files pending removal.
        """
        current = [item for item in self._files if item not in self._files_to_remove]
        current.extend(self._files_to_add)
        return list(sorted(set(current)))

    def load(self):
        """Load data payload and list of files."""
        self._payload = self.adapter.load_payload(self.id)
        self._payload_cache = copy.deepcopy(self._payload)

        self._files = self.adapter.list_files(self.id)
        self._files_to_add = {}
        self._files_to_remove = []

    def save(self):
        """Save data payload and attach/remove files."""
        if self.id is None:
            self.id = self.adapter.create_output(self.parent_id, payload=self.payload)
        else:
            self.adapter.save_payload(self.id, self.payload)

        for name in self._files_to_remove:
            self.adapter.remove_file(self.id, name)

        for name, path in self._files_to_add.items():
            with open(path, "rb") as infile:
                self.adapter.add_file(
                    self.id, name, original_name=path.name, content=infile.read()
                )

        # Empty unsaved values
        self._payload = self._payload_cache
        self._payload_cache = copy.deepcopy(self._payload)

        self._files = self.files
        self._files_to_add = {}
        self._files_to_remove = []

    def get_file(self, name, path=None) -> str:
        """Load an attached file and store it on the local filesystem.

        :param name: Name of attached file
        :param path: Destination path. Default to current working directory.
        :returns:    Path to created file
        """
        if name not in self.files:
            raise FileNotFoundError(f"No such file: {name}")

        if not path:
            root = os.getenv("ROBOT_ROOT", "")
            path = os.path.join(root, name)

        if name in self._files_to_add:
            local_path = self._files_to_add[name]
            if Path(local_path).resolve() != Path(path).resolve():
                copy2(local_path, path)
        else:
            content = self.adapter.get_file(self.id, name)
            with open(path, "wb") as outfile:
                outfile.write(content)

        # Always return absolute path
        return str(Path(path).resolve())

    def add_file(self, path, name=None):
        """Add file to current work item. Does not upload
        until ``save()`` is called.

        :param path: Path to file to upload
        :param name: Name of file in work item. If not given,
                     name of file on disk is used.
        """
        path = Path(path).resolve()

        if path in self._files_to_add.values():
            logging.warning("File already added: %s", path)

        if not path.is_file():
            raise FileNotFoundError(f"Not a valid file: {path}")

        name = name or path.name
        self._files_to_add[name] = path

        if name in self._files_to_remove:
            self._files_to_remove.remove(name)

        return name

    def remove_file(self, name, missing_ok=True):
        """Remove file from current work item. Change is not applied
        until ``save()`` is called.

        :param name: Name of attached file
        """
        if not missing_ok and name not in self.files:
            raise FileNotFoundError(f"No such file: {name}")

        if name in self._files:
            self._files_to_remove.append(name)

        if name in self._files_to_add:
            del self._files_to_add[name]

        return name


@library
class WorkItems:
    """A library for interacting with Control Room work items.

    Work items are used for managing data that go through multiple
    steps and tasks inside a process. Each step of a process receives
    input work items from the previous step, and creates output work items for
    the next step.

    **Item structure**

    A work item's data payload is JSON and allows storing anything that is
    serializable. This library by default interacts with payloads that
    are a dictionary of key-value pairs, which it treats as individual
    variables. These variables can be exposed to the Robot Framework task
    to be used directly.

    In addition to the data section, a work item can also contain files,
    which are stored by default in Robocorp Control Room. Adding and using
    files with work items requires no additional setup from the user.

    **Loading inputs**

    The library automatically loads the first input work item, if the
    library input argument ``autoload`` is truthy (default).

    After an input has been loaded its payload and files can be accessed
    through corresponding keywords, and optionally these values can be modified.

    **E-mail triggering**

    Since a process can be started in Control Room by sending an e-mail, a body
    in Text/JSON/YAML/HTML format can be sent as well and this gets attached to the
    input work item with the ``rawEmail`` payload variable. This library automatically
    parses the content of it and saves into ``parsedEmail`` the dictionary
    transformation of the original e-mail.

    If "Parse email" Control Room configuration option is enabled (recommended), then
    your e-mail is automatically parsed in the work item under the ``email`` payload
    variable, which is a dictionary containing a ``body`` holding the final parsed form
    of the interpreted e-mail body. The payload variable ``parsedEmail`` is still
    available for backwards compatibility reasons and holds the very same body inside
    the ``parsedEmail[Body]``.

    E-mail attachments will be added into the work item as files. Read more on:
    https://robocorp.com/docs/control-room/attended-or-unattended/email-trigger

    Example:

    After starting the process by sending an e-mail with a body like:

    .. code-block:: json

        {
            "message": "Hello world!"
        }

    The robot can use the parsed e-mail body's dictionary:

    .. code-block:: robotframework

        *** Tasks ***
        Using Parsed Emails
            ${mail} =    Get Work Item Variable    email
            Set Work Item Variables    &{mail}[body]
            ${message} =     Get Work Item Variable     message
            Log    ${message}    # will print "Hello world!"

    The behaviour can be disabled by loading the library with
    ``auto_parse_email=${None}`` or altered by providing to it a dictionary with one
    "key: value" where the key is usually "email.text" (deprecated "rawEmail", the
    variable set by Control Room, which acts as source for the parsed (deprecated raw)
    e-mail data) and the value can be "email.body" (deprecated "parsedEmail", where the
    parsed e-mail data gets stored into), value which can be customized and retrieved
    with ``Get Work Item Variable``.

    **Creating outputs**

    It's possible to create multiple new work items as an output from a
    task. With the keyword ``Create Output Work Item`` a new empty item
    is created as a child for the currently loaded input.

    All created output items are sent into the input queue of the next
    step in the process.

    **Active work item**

    Keywords that read or write from a work item always operate on the currently
    active work item. Usually that is the input item that has been automatically
    loaded when the execution started, but the currently active item is changed
    whenever the keywords ``Create Output Work Item`` or ``Get Input Work Item``
    are called. It's also possible to change the active item manually with the
    keyword ``Set current work item``.

    **Saving changes**

    While a work item is loaded automatically when a suite starts, changes are
    not automatically reflected back to the source. The work item will be modified
    locally and then saved when the keyword ``Save Work Item`` is called.
    This also applies to created output work items.

    It is recommended to defer saves until all changes have been made to prevent
    leaving work items in a half-modified state in case of failures.

    **Local Development**

    While Control Room is the default implementation, it can also be replaced
    with a custom adapter. The selection is based on either the ``default_adapter``
    argument for the library, or the ``RPA_WORKITEMS_ADAPTER`` environment
    variable. The library has a built-in alternative adapter called FileAdapter for
    storing work items to disk.

    The FileAdapter uses a local JSON file for input work items.
    It's a list of work items, each of which has a data payload and files.

    An example of a local file with one work item:

    .. code-block:: json

        [
            {
                "payload": {
                    "variable1": "a-string-value",
                    "variable2": ["a", "list", "value"]
                },
                "files": {
                    "file1": "path/to/file.ext"
                }
            }
        ]

    Output work items (if any) are saved to an adjacent file
    with the same name, but with the extension ``.output.json``. You can specify
    through the "RPA_OUTPUT_WORKITEM_PATH" env var a different path and name for this
    file.

    **Simulating the Cloud with Robocorp Code VSCode Extension**

    If you are developing in VSCode with the `Robocorp Code extension`_, you can
    utilize the built in local development features described in the
    `Developing with work items locally`_ section of the
    `Using work items`_ development guide.

    .. _Robocorp Code extension: https://robocorp.com/docs/setup/development-environment#visual-studio-code-with-robocorp-extensions
    .. _Developing with work items locally: https://robocorp.com/docs/development-guide/control-room/work-items#developing-with-work-items-locally
    .. _Using work items: https://robocorp.com/docs/development-guide/control-room/work-items

    **Examples**

    **Robot Framework**

    In the following example a task creates an output work item,
    and attaches some variables to it.

    .. code-block:: robotframework

        *** Settings ***
        Library    RPA.Robocorp.WorkItems

        *** Tasks ***
        Save variables to Control Room
            Create Output Work Item
            Set work item variables    user=Dude    mail=address@company.com
            Save Work Item

    In the next step of the process inside a different robot, we can use
    previously saved work item variables. Also note how the input work item is
    loaded implicitly when the suite starts.

    .. code-block:: robotframework

        *** Settings ***
        Library    RPA.Robocorp.WorkItems

        *** Tasks ***
        Use variables from Control Room
            Set task variables from work item
            Log    Variables are now available: s${user}, ${mail}

    **Python**

    The library can also be used through Python, but it does not implicitly
    load the first work item.

    .. code-block:: python

        import logging
        from RPA.Robocorp.WorkItems import WorkItems

        def list_variables(item_id):
            library = WorkItems()
            library.get_input_work_item()

            variables = library.get_work_item_variables()
            for variable, value in variables.items():
                logging.info("%s = %s", variable, value)
    """  # noqa: E501

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"
    ROBOT_LISTENER_API_VERSION = 2

    EMAIL_BODY_LOADERS = [
        ("JSON", json.loads),
        ("YAML", yaml.full_load),
    ]

    def __init__(
        self,
        autoload: bool = True,
        root: Optional[str] = None,
        default_adapter: Union[Type[BaseAdapter], str] = RobocorpAdapter,
        # pylint: disable=dangerous-default-value
        auto_parse_email: AUTO_PARSE_EMAIL_TYPE = AUTO_PARSE_EMAIL_DEFAULT,
    ):
        self.ROBOT_LIBRARY_LISTENER = self

        #: Current selected work item
        self._current: Optional[WorkItem] = None
        #: Input work items
        self.inputs: List[WorkItem] = []
        #: Output work items
        self.outputs: List[WorkItem] = []
        #: Variables root object in payload
        self.root = root
        #: Auto-load first input item and automatically parse e-mail content if
        # present.
        self.autoload: bool = autoload
        self._auto_parse_email = auto_parse_email
        #: Adapter for reading/writing items
        self._adapter_class = self._load_adapter(default_adapter)
        self._adapter: Optional[BaseAdapter] = None

        # Know when we're iterating (and consuming) all the work items in the queue.
        self._under_iteration = Event()

    @property
    def adapter(self):
        if self._adapter is None:
            self._adapter = self._adapter_class()
        return self._adapter

    @property
    def current(self) -> WorkItem:
        if self._current is None:
            raise RuntimeError("No active work item")

        return self._current

    @current.setter
    def current(self, value):
        if not isinstance(value, WorkItem):
            raise ValueError(f"Not a work item: {value}")

        self._current = value

    @property
    def active_input(self) -> Optional[WorkItem]:
        if self._current and self._current.parent_id is None:  # input set as current
            return self._current
        if self.inputs:  # other current item set, and taking the last input
            return self.inputs[-1]
        return None

    def _load_adapter(self, default) -> Type[BaseAdapter]:
        """Load adapter by name, using env or given default."""
        adapter = required_env("RPA_WORKITEMS_ADAPTER", default)

        if isinstance(adapter, str):
            adapter = import_by_name(adapter, __name__)

        assert issubclass(
            adapter, BaseAdapter
        ), "Adapter does not inherit from BaseAdapter"

        return adapter

    @classmethod
    def _interpret_content(cls, body: str) -> Union[dict, str]:
        for name, loader in cls.EMAIL_BODY_LOADERS:
            try:
                body = loader(body)
            except Exception as exc:  # pylint: disable=broad-except
                logging.debug(
                    "Failed deserializing input e-mail body content with loader %r "
                    "due to: %s",
                    name,
                    exc,
                )
            else:
                break

        return body

    def _get_email_content(
        self, variables: Dict
    ) -> Optional[Tuple[str, bool, Tuple[str]]]:
        # Returns the extracted e-mail [parsed] content and its payload destination.
        to_tuple = (
            lambda keys: (  # noqa: E731, pylint: disable=unnecessary-lambda-assignment
                keys if isinstance(keys, tuple) else (keys,)
            )
        )
        file_list = self.list_work_item_files()
        for input_keys, output_keys in self._auto_parse_email.items():
            input_keys = to_tuple(input_keys)
            for input_key in input_keys:
                try:
                    content = get_dot_value(variables, input_key)
                except AttributeError:
                    content = None  # couldn't reach final destination
                if not content and input_key in file_list:
                    path = Path(self.get_work_item_file(input_key))
                    content = path.read_text(encoding=ENCODING)
                    path.unlink()
                if content:
                    parsed = not input_key == "rawEmail"
                    if not parsed:
                        deprecation(
                            "Legacy non-parsed e-mail trigger detected! Please enable "
                            '"Parse email" configuration option in Control Room. (more'
                            " details: https://robocorp.com/docs/control-room/attended"
                            "-or-unattended/email-trigger#parse-email)"
                        )
                    output_keys = to_tuple(output_keys)
                    return content, parsed, output_keys
        return None

    def _parse_work_item_from_email(self):
        # Parse and return a dictionary from the input work item of a process started
        #  by e-mail trigger.
        if not self._auto_parse_email:
            return  # auto e-mail parsing disabled

        try:
            variables = self.get_work_item_variables()
        except ValueError:
            return  # payload not a dictionary

        content_details = self._get_email_content(variables)
        if not content_details:
            return  # no e-mail content found in the work item

        content, parsed, output_keys = content_details
        if parsed:  # With "Parse email" Control Room configuration option enabled.
            email_data = self._interpret_content(content)
        else:  # With "Parse email" Control Room configuration option disabled.
            # pylint: disable=no-member
            message = email.message_from_string(content)
            message_dict = dict(message.items())
            # pylint: disable=import-outside-toplevel
            from RPA.Email.ImapSmtp import ImapSmtp

            body, has_attachments = ImapSmtp().get_decoded_email_body(message)
            message_dict["Body"] = self._interpret_content(body)
            message_dict["Has-Attachments"] = has_attachments
            email_data = message_dict

        for output_key in output_keys:
            keys = output_key.split(".", 1)
            parsed_email = self.get_work_item_variable(keys[0], default={})
            if len(keys) == 2:
                set_dot_value(parsed_email, keys[1], value=email_data)
            else:
                parsed_email = email_data
            # pylint: disable=undefined-loop-variable
            self.set_work_item_variable(keys[0], parsed_email)

    def _start_suite(self, *_):
        """Robot Framework listener method, called when suite starts."""
        if not self.autoload:
            return

        try:
            self.get_input_work_item(_internal_call=True)
        # pylint: disable=broad-except
        except Exception as exc:
            logging.warning("Failed to load input work item: %s", exc)
        finally:
            self.autoload = False

    def _release_on_failure(self, attributes):
        """Automatically releases current input Work Item when encountering failures
        with tasks and/or suites.
        """
        if attributes["status"] != "FAIL":
            return

        message = attributes["message"]
        logging.info("Releasing FAILED input item with APPLICATION error: %s", message)
        self.release_input_work_item(
            state=State.FAILED,
            exception_type=Error.APPLICATION,
            message=message,
            _internal_release=True,
        )

    def _end_suite(self, _, attributes):
        """Robot Framework listener method, called when the suite ends."""
        # pylint: disable=unused-argument
        for item in self.inputs + self.outputs:
            if item.is_dirty:
                logging.warning("%s has unsaved changes that will be discarded", item)

        self._release_on_failure(attributes)

    def _end_test(self, _, attributes):
        """Robot Framework listener method, called when each task ends."""
        self._release_on_failure(attributes)

    @keyword
    def set_current_work_item(self, item: WorkItem):
        # pylint: disable=anomalous-backslash-in-string
        """Set the currently active work item.

        The current work item is used as the target by other keywords
        in this library.

        Keywords ``Get Input Work Item`` and ``Create Output Work Item``
        set the active work item automatically, and return the created
        instance.

        With this keyword the active work item can be set manually.

        Robot Framework Example:

        .. code-block:: robotframework

            *** Tasks ***
            Creating outputs
                ${input}=    Get Input Work Item
                ${output}=   Create Output Work Item
                Set current work item    ${input}

        Python Example:

        .. code-block:: python

            from RPA.Robocorp.WorkItems import WorkItems

            wi = WorkItems()
            parent_wi = wi.get_input_work_item()
            child_wi = wi.create_output_work_item()
            wi.set_current_work_item(parent_wi)
        """  # noqa: W605
        self.current = item

    @keyword
    def get_input_work_item(self, _internal_call: bool = False) -> WorkItem:
        """Load the next work item from the input queue, and set it as the active work
        item.

        Each time this is called, the previous input work item is released (as DONE)
        prior to reserving the next one.
        If the library import argument ``autoload`` is truthy (default),
        this is called automatically when the Robot Framework suite
        starts.
        """
        if not _internal_call:
            self._raise_under_iteration("Get Input Work Item")

        # Automatically release (with success) the lastly retrieved input work item
        # when asking for the next one. (or the currently set input if such)
        self.release_input_work_item(State.DONE, _internal_release=True)

        item_id = self.adapter.reserve_input()
        item = WorkItem(item_id=item_id, parent_id=None, adapter=self.adapter)
        item.load()
        self.inputs.append(item)
        self.current = item

        # Checks for raw/parsed e-mail content and parses it if present. This happens
        # with Processes triggered by e-mail.
        self._parse_work_item_from_email()

        return self.current

    @keyword
    def create_output_work_item(
        self,
        variables: Optional[dict] = None,
        files: Optional[Union[str, List[str]]] = None,
        save: bool = False,
    ) -> WorkItem:
        """Create a new output work item with optional variables and files.

        An output work item is always created as a child for an input item, therefore
        a non-released input is required to be loaded first.
        All changes to the work item are done locally and are sent to the output queue
        after the keyword ``Save Work Item`` is called only, except when `save` is
        `True`.

        :param variables: Optional dictionary with variables to be set into the new
            output work item.
        :param files: Optional list or comma separated paths to files to be included
            into the new output work item.
        :param save: Automatically call ``Save Work Item`` over the newly created
            output work item.
        :returns: The newly created output work item object.

        **Examples**

        **Robot Framework**

        .. code-block:: robotframework

            *** Tasks ***
            Create output items with variables then save
                ${customers} =  Load customer data
                FOR     ${customer}    IN    @{customers}
                    Create Output Work Item
                    Set Work Item Variables    id=${customer.id}
                    ...     name=${customer.name}
                    Save Work Item
                END

            Create and save output items with variables and files in one go
                ${customers} =  Load customer data
                FOR     ${customer}    IN    @{customers}
                    &{customer_vars} =    Create Dictionary    id=${customer.id}
                    ...     name=${customer.name}
                    Create Output Work Item     variables=${customer_vars}
                    ...     files=devdata${/}report.csv   save=${True}
                END

        **Python**

        .. code-block:: python

            from RPA.Robocorp.WorkItems import WorkItems

            wi = WorkItems()
            wi.get_input_work_item()
            customers = wi.get_work_item_variable("customers")
            for customer in customers:
                wi.create_output_work_item(customer, save=True)

        """
        if not self.inputs:
            raise RuntimeError(
                "Unable to create output work item without an input, "
                "call `Get Input Work Item` first"
            )

        parent = self.active_input
        if parent.state is not None:
            raise RuntimeError(
                "Can't create any more output work items since the last input was "
                "released, get a new input work item first"
            )

        item = WorkItem(item_id=None, parent_id=parent.id, adapter=self.adapter)
        self.outputs.append(item)
        self.current = item
        if variables:
            self.set_work_item_variables(**variables)
        if files:
            if isinstance(files, str):
                files = [path.strip() for path in files.split(",")]
            for path in files:
                # Assumes the name would be the same as the file name itself.
                self.add_work_item_file(path)
        if save:
            logging.debug("Auto-saving the just created output work item.")
            self.save_work_item()

        return self.current

    @keyword
    def save_work_item(self):
        """Save the current data and files in the work item. If not saved,
        all changes are discarded when the library goes out of scope.
        """
        self.current.save()

    @keyword
    def clear_work_item(self):
        """Remove all data and files in the current work item.

        Example:

        .. code-block:: robotframework

            *** Tasks ***
            Clearing a work item
                Clear work item
                Save work item

        .. code-block:: python

            from RPA.Robocorp.WorkItems import WorkItems

            wi = WorkItems()
            wi.get_input_work_item()
            wi.clear_work_item()
            wi.save_work_item()
        """
        self.current.payload.clear()
        self.remove_work_item_files("*")

    @keyword
    def get_work_item_payload(self):
        """Get the full JSON payload for a work item.

        **NOTE**: Most use cases should prefer higher-level keywords.

        Example:

        .. code-block:: robotframework

            *** Tasks ***
            Example task
                ${payload}=    Get work item payload
                Log    Entire payload as dictionary: ${payload}
        """
        return self.current.payload

    @keyword
    def set_work_item_payload(self, payload):
        # pylint: disable=anomalous-backslash-in-string
        """Set the full JSON payload for a work item.

        :param payload: Content of payload, must be JSON-serializable

        **NOTE**: Most use cases should prefer higher-level keywords.
        Using this keyword may cause errors when getting the payload via
        the normal ``Get work item variable`` and
        ``Get work item variables`` keywords if you do not set the payload
        to a ``dict``.

        Example:

        .. code-block:: robotframework

            *** Tasks ***
            Example task
                ${output}=    Create dictionary    url=example.com    username=Mark
                Set work item payload    ${output}

        """  # noqa: W605
        self.current.payload = payload

    @keyword
    def list_work_item_variables(self):
        """List the variable names for the current work item.

        Example:

        .. code-block:: robotframework

            *** Tasks ***
            Example task
                ${variables}=    List work item variables
                Log    Available variables in work item: ${variables}

        """
        return list(self.get_work_item_variables().keys())

    @keyword
    def get_work_item_variable(self, name, default=UNDEFINED):
        """Return a single variable value from the work item,
        or default value if defined and key does not exist.

        If key does not exist and default is not defined, raises `KeyError`.

        :param name: Name of variable
        :param default: Default value if key does not exist

        Robot Framework Example:

        .. code-block:: robotframework

            *** Tasks ***
            Using a work item
                ${username}=    Get work item variable    username    default=guest

        Python Example:

        .. code-block:: python

            from RPA.Robocorp.WorkItems import WorkItems

            wi = WorkItems()
            wi.get_input_work_item()
            customers = wi.get_work_item_variable("customers")
            print(customers)
        """
        variables = self.get_work_item_variables()
        value = variables.get(name, default)

        if value is UNDEFINED:
            raise KeyError(f"Undefined variable: {name}")

        notebook_print(text=f"**{name}** = **{value}**")
        return value

    @keyword
    def get_work_item_variables(self):
        """Read all variables from the current work item and
        return their names and values as a dictionary.

        Robot Framework Example:

        .. code-block:: robotframework

            *** Tasks ***
            Example task
                ${variables}=    Get work item variables
                Log    Username: ${variables}[username], Email: ${variables}[email]

        Python Example:

            from RPA.Robocorp.WorkItems import WorkItems
            wi = WorkItems()
            wi.get_input_work_item()
            input_wi = wi.get_work_item_variables()
            print(input_wi["username"])
            print(input_wi["email"])
        """

        payload = self.current.payload
        if not isinstance(payload, dict):
            raise ValueError(
                f"Expected work item payload to be `dict`, was `{type(payload)}`"
            )

        if self.root is not None:
            return payload.setdefault(self.root, {})

        return payload

    @keyword
    def set_work_item_variable(self, name, value):
        """Set a single variable value in the current work item.

        :param name: Name of variable
        :param value: Value of variable

        Robot Framework Example:

        .. code-block:: robotframework

            *** Tasks ***
            Example task
                Set work item variable    username    MarkyMark
                Save Work Item

        Python Example:

        .. code-block:: python

            from RPA.Robocorp.WorkItems import WorkItems

            customers = [{"id": 1, "name": "Apple"}, {"id": 2, "name": "Microsoft"}]
            wi = WorkItems()
            wi.get_input_work_item()
            wi.set_work_item_variable("customers", customers)
        """
        variables = self.get_work_item_variables()
        logging.info("%s = %s", name, value)
        variables[name] = value

    @keyword
    def set_work_item_variables(self, **kwargs):
        """Set multiple variables in the current work item.

        :param kwargs: Pairs of variable names and values

        Example:

        .. code-block:: robotframework

            *** Tasks ***
            Example task
                Set work item variables    username=MarkyMark    email=mark@example.com
                Save Work Item
        """
        variables = self.get_work_item_variables()
        for name, value in kwargs.items():
            logging.info("%s = %s", name, value)
            variables[name] = value

    @keyword
    def delete_work_item_variables(self, *names, force=True):
        """Delete variable(s) from the current work item.

        :param names: Names of variables to remove
        :param force: Ignore variables that don't exist in work item

        Example:

        .. code-block:: robotframework

            *** Tasks ***
            Example task
                Delete work item variables    username    email
                Save Work Item
        """
        variables = self.get_work_item_variables()
        for name in names:
            if name in variables:
                del variables[name]
                logging.info("Deleted variable: %s", name)
            elif not force:
                raise KeyError(f"No such variable: {name}")

    @keyword
    def set_task_variables_from_work_item(self):
        """Convert all variables in the current work item to
        Robot Framework task variables, see `variable scopes`_.

        .. _variable scopes: https://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#variable-scopes

        Example:

        .. code-block:: robotframework

            *** Tasks ***
            Example task
                # Work item has variable INPUT_URL
                Set task variables from work item
                Log    The variable is now available: ${INPUT_URL}
        """  # noqa: E501
        variables = self.get_work_item_variables()
        for name, value in variables.items():
            BuiltIn().set_task_variable(f"${{{name}}}", value)

    @keyword
    def list_work_item_files(self):
        """List the names of files attached to the current work item.

        Example:

        .. code-block:: robotframework

            *** Tasks ***
            Example task
                ${names}=    List work item files
                Log    Work item has files with names: ${names}
        """
        return self.current.files

    @keyword
    def get_work_item_file(self, name, path=None) -> str:
        """Get attached file from work item to disk.
        Returns the absolute path to the created file.

        :param name: Name of attached file
        :param path: Destination path of file. If not given, current
                     working directory is used.

        Example:

        .. code-block:: robotframework

            *** Tasks ***
            Example task
                ${path}=    Get work item file    input.xls
                Open workbook    ${path}
        """
        path = self.current.get_file(name, path)
        logging.info("Downloaded file to: %s", path)
        return path

    @keyword
    def add_work_item_file(self, path, name=None):
        """Add given file to work item.

        :param path: Path to file on disk
        :param name: Destination name for file. If not given, current name
                     of local file is used.

        **NOTE**: Files are not uploaded before work item is saved

        Example:

        .. code-block:: robotframework

            *** Tasks ***
            Example task
                Add work item file    output.xls
                Save Work Item
        """
        logging.info("Adding file: %s", path)
        return self.current.add_file(path, name=name)

    @keyword
    def remove_work_item_file(self, name, missing_ok=True):
        """Remove attached file from work item.

        :param name: Name of attached file
        :param missing_ok: Do not raise exception if file doesn't exist

        **NOTE**: Files are not deleted before work item is saved

        Example:

        .. code-block:: robotframework

            *** Tasks ***
            Example task
                Remove work item file    input.xls
                Save Work Item
        """
        logging.info("Removing file: %s", name)
        return self.current.remove_file(name, missing_ok)

    @keyword
    def get_work_item_files(self, pattern, dirname=None) -> List[str]:
        """Get files attached to work item that match given pattern.
        Returns a list of absolute paths to the downloaded files.

        :param pattern: Filename wildcard pattern
        :param dirname: Destination directory, if not given robot root is used

        Example:

        .. code-block:: robotframework

            *** Tasks ***
            Example task
                ${paths}=    Get work item files    customer_*.xlsx
                FOR  ${path}  IN  @{paths}
                    Handle customer file    ${path}
                END
        """
        paths = []
        for name in self.list_work_item_files():
            if fnmatch.fnmatch(name, pattern):
                if dirname:
                    path = self.get_work_item_file(name, os.path.join(dirname, name))
                else:
                    path = self.get_work_item_file(name)
                paths.append(path)

        logging.info("Downloaded %d file(s)", len(paths))
        return paths

    @keyword
    def add_work_item_files(self, pattern):
        """Add all files that match given pattern to work item.

        :param pattern: Path wildcard pattern

        Example:

        .. code-block:: robotframework

            *** Tasks ***
            Example task
                Add work item files    %{ROBOT_ROOT}/generated/*.csv
                Save Work Item
        """
        matches = FileSystem().find_files(pattern, include_dirs=False)

        paths = []
        for match in matches:
            path = self.add_work_item_file(match)
            paths.append(path)

        logging.info("Added %d file(s)", len(paths))
        return paths

    @keyword
    def remove_work_item_files(self, pattern, missing_ok=True):
        """Removes files attached to work item that match the given pattern.

        :param pattern: Filename wildcard pattern
        :param missing_ok: Do not raise exception if file doesn't exist

        Example:

        .. code-block:: robotframework

            *** Tasks ***
            Example task
                Remove work item files    *.xlsx
                Save Work Item
        """
        names = []

        for name in self.list_work_item_files():
            if fnmatch.fnmatch(name, pattern):
                name = self.remove_work_item_file(name, missing_ok)
                names.append(name)

        logging.info("Removed %d file(s)", len(names))
        return names

    def _raise_under_iteration(self, action: str) -> None:
        if self._under_iteration.is_set():
            raise RuntimeError(f"Can't {action} while iterating input work items")

    def _ensure_input_for_iteration(self) -> bool:
        active_input = self.active_input
        active_state = active_input.state if active_input else None
        if not active_input or active_state:
            # There are no inputs loaded yet or the lastly retrieved input work
            #  item is already processed. Time for trying to load a new one.
            try:
                self.get_input_work_item(_internal_call=True)
            except EmptyQueue:
                return False

        return True

    @keyword
    def for_each_input_work_item(
        self,
        keyword_or_func: Union[str, Callable],
        *args,
        items_limit: int = 0,
        return_results: bool = True,
        **kwargs,
    ) -> List[Any]:
        """Run a keyword or function for each work item in the input queue.

        Automatically collects and returns a list of results, switch
        ``return_results`` to ``False`` for avoiding this.

        :param keyword_or_func: The RF keyword or Py function you want to map through
            all the work items
        :param args: Variable list of arguments that go into the called keyword/function
        :param kwargs: Variable list of keyword arguments that go into the called
            keyword/function
        :param items_limit: Limit the queue item retrieval to a certain amount,
            otherwise all the items are retrieved from the queue until depletion
        :param return_results: Collect and return a list of results given each
            keyword/function call if truthy

        Example:

        .. code-block:: robotframework

            *** Tasks ***
            Log Payloads
                @{lengths} =     For Each Input Work Item    Log Payload
                Log   Payload lengths: @{lengths}

            *** Keywords ***
            Log Payload
                ${payload} =     Get Work Item Payload
                Log To Console    ${payload}
                ${len} =     Get Length    ${payload}
                [Return]    ${len}

        OR

        .. code-block:: python

            import logging
            from RPA.Robocorp.WorkItems import WorkItems

            library = WorkItems()

            def log_payload():
                payload = library.get_work_item_payload()
                print(payload)
                return len(payload)

            def log_payloads():
                library.get_input_work_item()
                lengths = library.for_each_input_work_item(log_payload)
                logging.info("Payload lengths: %s", lengths)

            log_payloads()
        """

        self._raise_under_iteration("iterate input work items")

        if isinstance(keyword_or_func, str):
            to_call = lambda: BuiltIn().run_keyword(  # noqa: E731,E501, pylint: disable=unnecessary-lambda-assignment
                keyword_or_func, *args, **kwargs
            )
        else:
            to_call = lambda: keyword_or_func(  # noqa: E731,E501, pylint: disable=unnecessary-lambda-assignment
                *args, **kwargs
            )
        results = []

        try:
            self._under_iteration.set()
            count = 0
            while True:
                input_ensured = self._ensure_input_for_iteration()
                if not input_ensured:
                    break

                result = to_call()
                if return_results:
                    results.append(result)
                self.release_input_work_item(State.DONE, _internal_release=True)
                count += 1
                if items_limit and count >= items_limit:
                    break
        finally:
            self._under_iteration.clear()

        return results if return_results else None

    @keyword
    def release_input_work_item(
        self,
        state: Union[State, str],
        exception_type: Optional[Union[Error, str]] = None,
        code: Optional[str] = None,
        message: Optional[str] = None,
        _internal_release: bool = False,
    ):
        """Release the lastly retrieved input work item and set its state.

        This can be released with DONE or FAILED states. With the FAILED state, an
        additional exception can be sent to Control Room describing the problem that
        you encountered by specifying a type and optionally a code and/or message.
        After this has been called, no more output work items can be created
        unless a new input work item has been loaded again.

        :param state: The status on the last processed input work item
        :param exception_type: Error type (BUSINESS, APPLICATION). If this is not
            specified, then the cloud will assume UNSPECIFIED
        :param code: Optional error code identifying the exception for future
            filtering, grouping and custom retrying behaviour in the cloud
        :param message: Optional human-friendly error message supplying additional
            details regarding the sent exception

        Example:

        .. code-block:: robotframework

            *** Tasks ***
            Example task
                Login into portal
                    ${user} =     Get Work Item Variable    user
                    ${doc} =    Get Work Item Variable    doc
                    TRY
                        Login Keyword    ${user}
                        Upload Doc Keyword    ${doc}

                    EXCEPT    Login Failed
                        Release Input Work Item     FAILED
                        ...    exception_type=APPLICATION
                        ...    code=LOGIN_PORTAL_DOWN
                        ...    message=Unable to login, retry again later.

                    EXCEPT    Format Error    AS    ${err}
                        ${message} =    Catenate
                        ...    Document format is not correct and cannot be uploaded.
                        ...    Correct the format in this work item and try again.
                        ...    Full error message received: ${err}
                        Release Input Work Item     FAILED
                        ...    exception_type=BUSINESS
                        ...    code=DOC_FORMAT_ERROR
                        ...    message=${message}

                    END

        OR

        .. code-block:: python

            from RPA.Robocorp.WorkItems import State, WorkItems

            library = WorkItems()

            def process_and_set_state():
                library.get_input_work_item()
                library.release_input_work_item(State.DONE)
                print(library.current.state)  # would print "State.DONE"

            process_and_set_state()
        """
        # Note that `_internal_release` here is True when automatically releasing items
        #  within our internal library logic.

        active_input = self.active_input
        if not active_input:
            if _internal_release:
                # Have nothing to release and that's normal (reserving for the first
                # time).
                return
            raise RuntimeError(
                "Can't release without reserving first an input work item"
            )
        if active_input.state is not None:
            if _internal_release:
                # Item already released and that's normal when reaching an empty queue
                # and we ask for another item again. We don't want to set states twice.
                return
            raise RuntimeError("Input work item already released")
        assert active_input.parent_id is None, "set state on output item"
        assert active_input.id is not None, "set state on input item with null ID"

        # RF automatically converts string "DONE" to State.DONE object if only `State`
        #  type annotation is used in the keyword definition.
        if not isinstance(state, State):
            # But since we support strings as well now, to stay compatible with Python
            #  behaviour, a "COMPLETE" value is expected instead of "DONE".
            state: str = state.upper()
            state: str = State.DONE.value if state == "DONE" else state
            state: State = State(state)
        exception = None
        if state is State.FAILED:
            if exception_type:
                exception_type: Error = (
                    exception_type
                    if isinstance(exception_type, Error)
                    else Error(exception_type.upper())
                )
                exception = {
                    "type": exception_type.value,
                    "code": code,
                    "message": message,
                }
            elif code or message:
                exc_types = ", ".join(list(Error.__members__))
                raise RuntimeError(f"Must specify failure type from: {exc_types}")

        self.adapter.release_input(active_input.id, state, exception=exception)
        active_input.state = state

    @keyword
    def get_current_work_item(self) -> WorkItem:
        """Get the currently active work item.

        The current work item is used as the target by other keywords
        in this library.

        Keywords ``Get Input Work Item`` and ``Create Output Work Item``
        set the active work item automatically, and return the created
        instance.

        With this keyword the active work item can be retrieved manually.

        Example:

        .. code-block:: robotframework

            *** Tasks ***
            Example task
                ${input} =    Get Current Work Item
                ${output} =   Create Output Work Item
                Set Current Work Item    ${input}
        """
        return self.current



import atexit
from dataclasses import dataclass
import logging
import os
import platform
import queue
import threading
import time
from typing import List, Optional, Union
import warnings

from robot.api.deco import library, keyword
from robot.libraries.BuiltIn import BuiltIn

from RPA.Desktop import Desktop
from RPA.core.logger import deprecation


DESKTOP = Desktop()


def get_scaled_coordinate(coordinate, scaling_factor):
    return int(coordinate * scaling_factor)


if platform.system() == "Windows":  # noqa: C901
    import ctypes
    from ctypes import wintypes, byref
    from JABWrapper.context_tree import ContextTree, ContextNode, SearchElement
    from JABWrapper.jab_wrapper import JavaAccessBridgeWrapper

    # Configure comtypes to not generate DLL bindings into
    # current environment, instead keeping them in memory.
    # Slower, but prevents dirtying environments.
    import comtypes.client

    comtypes.client.gen_dir = None

    # Ignore warning about threading mode,
    # which comtypes initializes to STA instead of MTA on import.
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=UserWarning)
        import win32con
        import win32gui

    PeekMessage = ctypes.windll.user32.PeekMessageW
    GetMessage = ctypes.windll.user32.GetMessageW
    TranslateMessage = ctypes.windll.user32.TranslateMessage
    DispatchMessage = ctypes.windll.user32.DispatchMessageW
    ScalingFactor: float = ctypes.windll.shcore.GetScaleFactorForDevice(0) / 100

    @dataclass
    class JavaElement:
        """Abstraction for Java object properties"""

        name: str
        role: str
        description: str
        states: list
        ancestry: int
        checked: bool
        selected: bool
        visible: bool
        enabled: bool
        showing: bool
        focusable: bool
        states_string: str
        x: int
        y: int
        width: int
        height: int
        node: ContextNode
        row: int
        col: int
        text: str
        column_count: int
        visible_children: list
        visible_children_count: int
        index_in_parent: int

        def __init__(
            self,
            node,
            scaling_factor: Optional[float] = None,
            index: int = 0,
            column_count: Optional[int] = None,
        ):
            scaling_factor = scaling_factor or ScalingFactor
            self.name = node.context_info.name
            self.role = node.context_info.role
            self.description = node.context_info.description
            self.states = node.context_info.states.split(",")
            self.ancestry = node.ancestry
            self.checked = "checked" in self.states
            self.selected = "selected" in self.states
            self.visible = "visible" in self.states
            self.enabled = "enabled" in self.states
            self.showing = "showing" in self.states
            self.focusable = "focusable" in self.states
            self.node = node
            self.states_string = node.context_info.states
            self.x = get_scaled_coordinate(node.context_info.x, scaling_factor)
            self.y = get_scaled_coordinate(node.context_info.y, scaling_factor)
            self.width = get_scaled_coordinate(node.context_info.width, scaling_factor)
            self.height = get_scaled_coordinate(
                node.context_info.height, scaling_factor
            )
            self.center_x = self.x + int(self.width / 2)
            self.center_y = self.y + int(self.height / 2)
            self.text = node.text._items.sentence
            self.visible_children = node.get_visible_children()
            self.visible_children_count = node.visible_children_count
            self.index_in_parent = node.context_info.indexInParent
            self.column_count = column_count
            self.set_rows_columns(self.visible_children, index=index)

        def set_rows_columns(self, children: List[ContextNode], index: int = 0):
            if not self.column_count:
                self.column_count = 0
                if children:
                    self.column_count = 1
                    java_child = JavaElement(children[0])
                    x_pos, y_pos = java_child.x, java_child.y
                    for child in children[1:]:
                        java_child = JavaElement(child)
                        if java_child.x == x_pos or java_child.y != y_pos:
                            break
                        self.column_count += 1

            if self.column_count > 0:
                self.row = (
                    0 if index < self.column_count else int(index / self.column_count)
                )
                self.col = (
                    index
                    if index < self.column_count
                    else int(index - (self.row * self.column_count))
                )
            else:
                self.row = -1
                self.col = -1

        def click(self, click_type: str = "click"):
            if self.x != -1 and self.y != -1:
                locator = f"coordinates:{self.center_x},{self.center_y}"
                DESKTOP.click(locator, action=click_type)

        def type_text(self, text: str, clear: bool = False) -> None:
            self.click()
            if clear:
                DESKTOP.press_keys("ctrl", "a")
                time.sleep(0.2)
                DESKTOP.press_keys("delete")
                time.sleep(0.2)
            for c in text:
                DESKTOP.press_keys(c)

        def __str__(self):
            string = (
                f"name={self.name}; "
                f"role={self.role}; "
                f"enabled={self.enabled}; "
                f"visible={self.visible}; "
                f"selected={self.selected}; "
                f"checked={self.checked}; "
                f"showing={self.checked}; "
                f"focusable={self.checked}; "
                f"description={self.description}; "
                f"ancestry={self.ancestry}; "
                f"states={self.states}; "
                f"x={self.x}; "
                f"y={self.y}; "
                f"width={self.width}; "
                f"height={self.height}; "
                f"center_x={self.center_x}; "
                f"center_y={self.center_y}; "
                f"indexInParent={self.index_in_parent}; "
                f"childrentCount={self.visible_children_count};"
            )
            return string

    LocatorType = Union[ContextNode, JavaElement, str]
else:
    ScalingFactor = 1.0
    LocatorType = str
    ContextNode = object
    JavaElement = object


class ElementNotFound(ValueError):
    """No matching elements were found."""


class JavaWindowNotFound(ValueError):
    """No active Java window were found."""


class InvalidLocatorError(AttributeError):
    """Invalid locator string."""


class JavaAccessBridgeError(RuntimeError):
    """Base exception for Java Access Bridge errors."""


IntegerLocatorTypes = [
    "x",
    "y",
    "width",
    "height",
    "indexInParent",
    "ancestry",
    "childrentCount",
]


@library(scope="GLOBAL", doc_format="REST", auto_keywords=False)
class JavaAccessBridge:
    # pylint: disable=W1401
    """Java application UI automation library using `Java Access Bridge technology`_.

    The library utilizes `java-access-bridge-wrapper`_ package to interact with
    Java UI. Currently only the 64-bit Windows OS is supported.

    **Inspecting elements**

    We have built an `Assistant`_ for working with Java application's element structure and `Java locators`_.
    The Assistant provides copy-paste-able locators for each element and also allows testing locators against
    selected application.

    If our tools fail to pick the locator from your target application, there is always the
    `Access Bridge Explorer`_ from Google that enables you to see the raw view. Please note that
    Access Bridge Explorer repository has been archived on July 27, 2022 and is no longer actively
    maintained.

    The `Accessibility Insights for Windows`_ can show element properties if application framework
    supports Windows UI Automation (UIA), see more at `using Accessibility Insights`_. Then the recommended
    library would be `RPA.Windows`_ library.

    **Steps to enable**

        1. Enable the Java Access Bridge in Windows
        2. Set environment variable `RC_JAVA_ACCESS_BRIDGE_DLL` as an absolute path to `WindowsAccessBridge-64.dll`.
           It is also possible to give DLL location as library initialization parameter `access_bridge_path`.

        .. code-block:: console

            C:\\path\\to\\java\\bin\\jabswitch -enable
            set RC_JAVA_ACCESS_BRIDGE_DLL=C:\\path\\to\\Java\\bin\\WindowsAccessBridge-64.dll

        .. code:: robotframework

            *** Settings ***
            Library   RPA.JavaAccessBridge   access_bridge_path=C:\\path\\to\\Java\\bin\\WindowsAccessBridge-64.dll

    .. _Java Access Bridge technology: https://www.oracle.com/java/technologies/javase/javase-tech-access-bridge.html
    .. _java-access-bridge-wrapper: https://github.com/robocorp/java-access-bridge-wrapper
    .. _Assistant: https://github.com/robocorp/working-with-java

    **About Java wrapper callbacks and actions**

    There might be a compatibility issue with callbacks and actions on target Java application. Possible reasons:

    - target application is executed with 32-bit Java
    - target application does not support callbacks and/or actions

    Workaround for this situation is to initialize `JavaAccessBridge` library with parameter `ignore_callbacks=True`.
    Then application's element information is still accessible and any actions on those elements can be performed
    with `RPA.Desktop` library. Keep in mind that you can still manuall refresh an element with ``Refresh Element``.

    *Note.* There are still keywords, for example. ``Call Element Action``, which will cause error if used in this
    situation.

    .. code:: robotframework

        *** Settings ***
        Library   RPA.JavaAccessBridge   ignore_callbacks=True

    **Controlling the Java window**

    Keyword for this purpose is ``Select Window``. Window selection is based on the ``title`` parameter, which can be
    given as a regular expressions to match the correct window. The keyword brings the window into focus and initially
    reads window's element structure.

    **Locating elements**

    To automate actions on the Java application, the robot needs locations to various elements
    using a feature called `Java locators`_. Locator describes properties of an element.

    At the moment library contains basic level support for locators.

    The common locator types are ``name`` and ``role``.

    To identify element with more than one property `and` can be used, for example:

        .. code-block:: console

            role:push button and name:Clear

    To address element within parent element `>` can be used, for example:

        .. code-block:: console

            name:Find Purchase Orders > name:NumberField

    Some keywords accept element as an parameter in place of locator.

    New locator type `strict` has been added in rpaframework==12.5.0. Currently
    property values of string type have been evaluated with `startsWith` which
    can match several property values. With `strict` set in the locator string,
    all locator on the right side of this definition will be matched using
    strict (equal matching), example:

        .. code-block:: robotframework

            # without strict, name can be 'Type', 'Type1', 'Type of'...
            Get Elements   role:push button and name:Type
            # name must be equal to 'Type'
            Get Elements  role:push button and strict:True and name:Type

    Keyword ``Get Elements`` has extra parameter ``strict``, which when set to
    ``True`` forces all locator value matches to be strict, example:

        .. code-block:: robotframework

            # without strict, name can be 'Type', 'Type1', 'Type of'...
            Get Elements  role:push button and name:Type
            # name must be equal to 'Type' and role must be equal to 'text'
            Get Elements  role:text and name:Type  strict=True

    **About JavaElement object**

    The ``JavaElement`` was added in rpaframework==12.3.0 for easy access into
    ``ContextNode`` objects which have been returned by ``Get Elements`` keyword.

    Keyword ``Get Elements`` still returns ``ContextNode`` objects, but with parameter
    ``java_elements=True`` the keyword returns ``JavaElement`` objects instead (they
    still contain reference to ``ContextNode`` object via ``node`` property, e.g.
    JavaObject.node).

    Properties and methods included in the JavaElement:

    - name: str
    - role: str
    - description: str
    - states: list            # list presentation of states (string)
    - ancestry: int           # you can set the maximum depth based on this
    - checked: bool
    - selected: bool
    - visible: bool
    - enabled: bool
    - showing: bool
    - focusable: bool
    - states_string: str
    - x: int                  # left coordinate of the element
    - y: int                  # top coordinate of the element
    - width: int
    - height: int
    - node: ContextNode       # original ContextNode
    - row: int                # table row, -1 if element is not member of table
    - col: int                # table column, -1 if element is not member of table
    - text: str               # text content of the element
    - column_count: int       # table column count
    - visible_children: list  # visible children elements of this element
    - visible_children_count: int
    - index_in_parent: int    # position in the parent
    - click()                 # method for clicking element center
    - type_text()             # method for typing text into element (if possible)

    **Interacting with elements**

    By default application elements are interacted with Actions supported by the element.
    Most common example is `click` action supported by an button element.

    But because application and technology support for the actions might be limited, it is also
    possible to opt for interaction elements by their coordinates by giving keyword parameter
    ``action=False`` if parameter is available.

    .. _Accessibility Insights for Windows: https://accessibilityinsights.io/en/downloads/
    .. _Access Bridge Explorer: https://github.com/google/access-bridge-explorer
    .. _using Accessibility Insights: https://accessibilityinsights.io/docs/en/windows/reference/faq/#can-i-use-accessibility-insights-for-windows-on-a-windows-app-written-with-java
    .. _RPA.Windows: https://robocorp.com/docs/libraries/rpa-framework/rpa-windows
    .. _Java locators: https://robocorp.com/docs/development-guide/locators#java-locators

    **Examples**

    *robotframework*

    .. code:: robotframework

        *** Settings ***
        Library   RPA.JavaAccessBridge
        Library   Process

        *** Tasks ***
        Write text into Swing application
            Start Process    java -jar BasicSwing.jar
            ...              shell=${TRUE}
            ...              cwd=${CURDIR}
            Select Window    Chat Frame
            Type Text    role:text
            ...          text for the textarea
            Type Text    role:text
            ...          text for the input field
            ...          index=1
            ...          clear=${TRUE}
            Click Element    role:push button and name:Send

    *Python*

    .. code:: python

        from RPA.JavaAccessBridge import JavaAccessBridge
        import subprocess

        jab = JavaAccessBridge()

        subprocess.Popen(
            ["java", "-jar", "BasicSwing.jar"],
            shell=True,
            cwd=".",
            close_fds=True
        )
        jab.select_window("Chat Frame")
        jab.type_text(
            "role:text",
            "text for the textarea",
            enter=True
        )
        jab.type_text(
            "role:text",
            "text for the input field",
            index=1,
            clear=True
        )
        jab.click_element("role:push button and name:Send")

    """  # noqa: E501, W605

    # TODO: add keyword for taking screenshots of elements and window
    # TODO. implement proper XPath syntax support

    def __init__(
        self,
        ignore_callbacks: bool = False,
        access_bridge_path: Optional[str] = None,
        max_depth: Optional[int] = None,
        disable_refresh: bool = False,
    ):
        """If library is not given ``access_bridge_path`` then path needs to be given
        by the environment variable ``RC_JAVA_ACCESS_BRIDGE_DLL``.

        The ``ignore_callbacks`` can be set to ``True`` if the target application
        does not support Java callback feature (property value change listeners).

        :param ignore_callbacks: set to `True` if application does not support
            Java callback feature, defaults to False
        :param access_bridge_path: absolute filepath to the DLL, defaults to None
        :param max_depth: limit the height of the element tree (how deep the search
            goes); by default every element is taken into account and the tree of
            elements is built until reaching the last leaf (which might be
            time-consuming)
        :param disable_refresh: disables automatic app/element refresh when this is
            `True` (saves time); by default, automatic app/element refreshes are
            triggered with keywords like ``Read Table``, ``Wait Until Element Exists``
        """
        self.logger = logging.getLogger(__name__)
        desktoplogger = logging.getLogger("RPA.Desktop")
        desktoplogger.setLevel(logging.WARNING)
        if platform.system() != "Windows":
            self.logger.warning(
                "JavaAccessBridge library requires Windows dependencies to work"
            )
        if access_bridge_path:
            os.environ["RC_JAVA_ACCESS_BRIDGE_DLL"] = access_bridge_path
        if "RC_JAVA_ACCESS_BRIDGE_DLL" not in os.environ.keys():
            self.logger.warning(
                "Environment variable `RC_JAVA_ACCESS_BRIDGE_DLL` needs to be set to "
                "absolute path of `WindowsAccessBridge-64.dll`"
            )
        self.version_printed = False
        self.jab_wrapper = None
        self.context_info_tree = None
        self.pumper_thread = None
        self.display_scale_factor = ScalingFactor
        self.ignore_callbacks = ignore_callbacks
        self.max_depth = max_depth
        self.disable_refresh = disable_refresh
        self.pid = None

    def _initialize(self):
        pipe = queue.Queue()
        self.pumper_thread = threading.Thread(
            target=self._pump_background, daemon=True, args=[pipe]
        )
        self.pumper_thread.start()
        obj = pipe.get(timeout=10)
        if isinstance(obj, Exception):
            raise obj
        if isinstance(obj, JavaAccessBridgeWrapper):
            self.jab_wrapper = obj
        if not self.jab_wrapper:
            raise JavaAccessBridgeError(
                "Failed to initialize Java Access Bridge Wrapper"
            )
        time.sleep(1)
        atexit.register(self._handle_shutdown)
        self.logger.info("Java Access Bridge Wrapper initialized")

    def _handle_shutdown(self):
        if self.jab_wrapper:
            self.jab_wrapper.shutdown()

    def _pump_background(self, pipe: queue.Queue):
        try:
            jab_wrapper = JavaAccessBridgeWrapper(
                ignore_callbacks=self.ignore_callbacks
            )
            pipe.put(jab_wrapper)
            message = byref(wintypes.MSG())
            while GetMessage(message, 0, 0, 0) > 0:
                TranslateMessage(message)
                self.logger.debug("Dispatching msg=%s", repr(message))
                DispatchMessage(message)
        # pylint: disable=broad-except
        except Exception as err:
            self.logger.error(err)
            pipe.put(err)
        finally:
            self.logger.info("Stopped processing events")

    @keyword
    def set_display_scale_factor(self, factor: float) -> float:
        """Override library display scale factor.

        Keyword returns previous value.

        :param factor: value for the new display scale factor
        :return: previous display scale factor value
        """
        previous_factor = self.display_scale_factor
        self.display_scale_factor = factor
        return previous_factor

    @keyword
    def select_window(
        self, title: str, bring_foreground: bool = True, timeout: int = 30
    ):
        """Selects Java application window as target for the automation using
        Java window title.

        :param title: application window title
        :param bring_foreground: if application is brought to foreground or not
        :param timeout: selection timeout
        """
        deprecation(
            "Keyword `Select Window` is deprecated and will be removed in the "
            "next major update to the `RPA.JavaAccessBridge` library. "
            "Please use `Select Window By Title` or `Select Window By PID` keywords "
            "instead."
        )
        self._select_window(
            title=title, bring_foreground=bring_foreground, timeout=timeout
        )

    def _parse_locator(self, locator, strict_default=False):
        levels = locator.split(">")
        levels = [lvl.strip() for lvl in levels]
        searches = []
        for lvl in levels:
            conditions = lvl.split(" and ")
            lvl_search = []
            strict_mode = strict_default
            for cond in conditions:
                parts = cond.split(":", 1)
                if len(parts) == 1:
                    parts = ["name", parts[0]]
                elif parts[0].lower() == "strict":
                    strict_mode = bool(parts[1])
                    continue
                elif parts[0] in IntegerLocatorTypes:
                    try:
                        parts[1] = int(parts[1])
                    except ValueError as err:
                        raise InvalidLocatorError(
                            "Locator '%s' needs to be of 'integer' type" % parts[0]
                        ) from err
                lvl_search.append(SearchElement(parts[0], parts[1], strict=strict_mode))
            searches.append(lvl_search)
        return searches

    def _find_elements(
        self, locator: str, index: Optional[int] = None, strict: bool = False
    ) -> Union[ContextNode, List[ContextNode]]:
        if not self.context_info_tree:
            raise ValueError("ContextTree has not been initialized")
        searches = self._parse_locator(locator, strict)
        self.logger.info("Searches: %s", searches)
        elements = []
        for lvl, search_elements in enumerate(searches):
            if lvl == 0:
                elements = self.context_info_tree.get_by_attrs(search_elements)
            else:
                sub_matches = []
                for elem in elements:
                    matches = elem.get_by_attrs(search_elements)
                    sub_matches.extend(matches)
                elements = sub_matches
        self.logger.info('Search "%s" returned %s element(s)', locator, len(elements))
        if index and len(elements) > (index + 1):
            raise AttributeError(
                "Locator '%s' returned only %s elements (can't index element at %s)"
                % (locator, len(elements), index)
            )
        return elements[index] if index else elements

    @keyword
    def set_mouse_position(self, element: ContextNode):
        """Set mouse position to element center

        :param element: target element
        """
        left, top, right, bottom = self._get_scaled_coordinates(element)
        middle_x = int((left + right) / 2)
        middle_y = int((top + bottom) / 2)
        point = f"point:{middle_x},{middle_y}"
        DESKTOP.move_mouse(point)

    @keyword
    def type_text(
        self,
        locator: LocatorType,
        text: str,
        index: int = 0,
        clear: bool = False,
        enter: bool = False,
        typing: bool = True,
    ):
        """Type text into coordinates defined by locator

        :param locator: target element
        :param text: text to write
        :param index: target element if multiple are returned
        :param clear: should element be cleared before typing
        :param enter: should enter key be pressed after typing
        :param typing: if True (default) will use Desktop().type_text()
         if False will use Desktop().press_keys()
        """
        target = self._get_matching_element(locator, index)
        self.click_element(target, click_type="double click")

        if not self.ignore_callbacks:
            target.request_focus()
            time.sleep(1)  # this takes time and interferes with the keys sending below
        if clear:
            DESKTOP.press_keys("ctrl", "a")
            time.sleep(0.2)
            DESKTOP.press_keys("delete")
            time.sleep(0.2)
        self.logger.info("type text: %s", text)
        if typing:
            DESKTOP.type_text(text, enter=enter)
        else:
            for c in text:
                DESKTOP.press_keys(c)
            if enter:
                DESKTOP.press_keys("enter")

    @keyword
    def get_elements(
        self, locator: str, java_elements: bool = False, strict: bool = False
    ) -> List[Union[ContextNode, JavaElement]]:
        """Get matching elements

        :param locator: elements to get
        :param java_elements: if True will return elements as ``JavaElement``
         on False will return Java ContextNodes
        :param strict: on True all locator matches need to match exactly, on
         False will be using startsWith matching on non-integer properties
        :return: list of ContextNodes or JavaElements

        Python example.

        .. code:: python

            elements = java.get_elements("name:common", java_elements=True)
            for e in elements:
                print(e.name if e.name else "EMPTY", e.visible, e.x, e.y)
                if e.role == "check box":
                    e.click()
                else:
                    java.type_text(e, "new content", clear=True, typing=False)

            # following does NOT return anything because search is strict
            # and there are no 'push butto' role
            elements = java.get_elements("role:push butto", strict=True)

        Robotframework  example.

        .. code:: robotframework

            ${elements}=    Get Elements
            ...    role:push button and name:Send
            ...    java_elements=True
            Evaluate   $elements[0].click()
            Click Element    ${elements}[0]    action=False
            Type Text
            ...    ${elements}[0]
            ...    moretext
            ...    clear=True
            ...    typing=False
        """
        elements = self._find_elements(locator, strict=strict)
        return (
            [JavaElement(e, self.display_scale_factor) for e in elements]
            if java_elements
            else elements
        )

    @keyword
    def wait_until_element_text_contains(
        self, locator: LocatorType, text: str, index: int = 0, timeout: float = 0.5
    ):
        """Wait until element text contains expected text

        :param locator: target element
        :param text: element text should contain this
        :param index: target element index if multiple are returned
        :param timeout: timeout in seconds to wait, default 0.5 seconds
        """
        matching = self._get_matching_element(locator, index)
        end_time = time.time() + float(timeout)
        while time.time() <= end_time:
            # pylint: disable=protected-access
            if text in matching.text._items.sentence:
                return
            time.sleep(0.05)

        raise ValueError(f"Text={text} not found in element={matching}")

    @keyword
    def wait_until_element_text_equals(
        self, locator: LocatorType, text: str, index: int = 0, timeout: float = 0.5
    ):
        """Wait until element text equals expected text

        :param locator: target element
        :param text: element text should match this
        :param index: target element index if multiple are returned
        :param timeout: timeout in seconds to wait, default 0.5 seconds
        """
        matching = self._get_matching_element(locator, index)
        end_time = time.time() + float(timeout)
        while time.time() <= end_time:
            # pylint: disable=protected-access
            if text == matching.text._items.sentence:
                return
            time.sleep(0.05)

        raise ValueError(f"Text={text} not found in element={matching}")

    @keyword
    def wait_until_element_is_focused(
        self, locator: LocatorType, index: int = 0, timeout: float = 0.5
    ):
        """Wait until element is focused

        :param locator: target element
        :param index: target element index if multiple are returned
        :param timeout: timeout in seconds to wait, default 0.5 seconds
        """
        java_element = self._get_matching_element(locator, index, as_java_element=True)
        end_time = time.time() + float(timeout)
        self.logger.info("Focusing on element: %s", java_element)

        while time.time() <= end_time:
            if "focused" in java_element.states:
                return
            time.sleep(0.05)

        raise ValueError(f"Element={java_element} not focused")

    @keyword
    def get_element_text(self, locator: LocatorType, index: int = 0):
        """Get element text

        :param locator: target element
        :param index: target element index if multiple are returned
        """
        matching = self._get_matching_element(locator, index)
        # pylint: disable=protected-access
        return matching.text._items.sentence

    def _get_matching_element(
        self,
        locator: LocatorType,
        index: int = 0,
        as_java_element: bool = False,
        refresh: bool = False,
    ) -> Union[ContextNode, JavaElement]:
        matching = None
        if isinstance(locator, str):
            elements = self._find_elements(locator)
            if len(elements) < (index + 1):
                raise ElementNotFound(
                    "Locator '%s' matched  %s elements" % (locator, len(elements))
                )
            matching = elements[index]
        elif isinstance(locator, ContextNode):
            matching = locator
        elif isinstance(locator, JavaElement):
            matching = locator.node

        if refresh and not self.disable_refresh:
            matching.refresh()
        return (
            JavaElement(matching, self.display_scale_factor)
            if as_java_element
            else matching
        )

    @keyword
    def get_element_actions(self, locator: LocatorType) -> List[str]:
        """Get list of possible element actions

        :param locator: target element
        """
        if isinstance(locator, str):
            target = self._find_elements(locator, index=0)
        elif isinstance(locator, ContextNode):
            target = locator
        else:
            target = locator.node
        return target.get_actions()

    def _elements_to_console(self, elements, function=""):
        BuiltIn().log_to_console(f"\nElements to Console: {function}")
        for elem in elements:
            BuiltIn().log_to_console(str(elem).strip())

    @keyword
    def highlight_element(self, locator: LocatorType, index: int = 0):
        """Highlight an element

        :param locator: element to highlight
        :param index: target element index if multiple are returned
        """
        matching = self._get_matching_element(locator, index)
        self.logger.info("Highlighting element: %s", repr(matching))
        region_locator = self._get_region_locator(matching)
        DESKTOP.highlight_elements(region_locator)

    def _get_scaled_coordinates(self, element):
        left = int(element.context_info.x / self.display_scale_factor)
        top = int(element.context_info.y / self.display_scale_factor)
        width = int(element.context_info.width / self.display_scale_factor)
        height = int(element.context_info.height / self.display_scale_factor)
        right = left + width
        bottom = top + height
        return left, top, right, bottom

    def _get_region_locator(self, element):
        left, top, right, bottom = self._get_scaled_coordinates(element)
        return DESKTOP.define_region(left, top, right, bottom)

    @keyword
    def click_element(
        self,
        locator: LocatorType,
        index: int = 0,
        action: bool = True,
        timeout: int = 10,
        click_type: str = "click",
    ):
        """Click element

        :param locator: element to click
        :param index: target element index if multiple are returned
        :param action: call click action on element (default), or use coordinates
        :param timeout: timeout in seconds to find element
        :param click_type: default `click`, see `RPA.Desktop` for different
         click options
        """
        if isinstance(locator, str):
            elements = self.wait_until_element_exists(locator, timeout)
            if len(elements) < (index + 1):
                # print element tree when result is not expected
                self.print_element_tree()
                raise ElementNotFound(
                    "Locator '%s' matched only %s elements" % (locator, len(elements))
                )
            element = elements[index]
        else:
            element = locator
        try:
            if action:
                self.logger.info("Element click action type: %s", type(element))
                element.click()  # works for both `ContextNode` & `JavaElement`
            else:
                self._click_element_middle(element, click_type)
        except NotImplementedError:
            self._click_element_middle(element, click_type)

    @keyword
    def call_element_action(self, locator: str, action: str):
        """Call element action

        :param locator: target element
        :param action: name of the element action to call
        """
        if isinstance(locator, str):
            elements = self._find_elements(locator)
            if len(elements) != 1:
                raise ElementNotFound(
                    "Locator %s did not match a unique element" % locator
                )

            matching = elements[0]
        else:
            matching = locator
        self.logger.info("Element '%s' action", action)
        matching.do_action(action)

    def _click_element_middle(self, element, click_type="click"):
        self.logger.info("Element click coordinates")
        if not isinstance(element, JavaElement):
            java_element = JavaElement(element, self.display_scale_factor)
        else:
            java_element = element
        self.click_coordinates(java_element.center_x, java_element.center_y, click_type)

    @keyword
    def click_coordinates(
        self, x: int, y: int, click_type: str = "click", delay: float = 0.5
    ):
        """Keyword to mouse click at specific coordinates.

        :param x: horizontal coordinate
        :param y: vertical coordinates
        :param click_type: default `click`, see `RPA.Desktop` for different
         click options
        :param delay: how much in seconds to delay after click, defaults to 0.5
        """
        locator = f"coordinates:{x},{y}"
        DESKTOP.click(locator, action=click_type)
        time.sleep(delay)

    @keyword
    def toggle_drop_down(self, locator: LocatorType, index: int = 0):
        """Toggle dropdown action on element

        :param locator: element locator
        :param index: target element index if multiple are returned
        """
        matching = self._get_matching_element(locator, index)
        matching.toggle_drop_down()

    def _application_refresh(self, force: bool = False):
        if self.disable_refresh and not force:
            self.logger.debug("Refresh triggered, but disabled globally by user.")
            return

        self.logger.info(
            "%s refreshing the entire app to a maximum depth of %d.",
            "Forcibly" if force else "Automatically",
            self.max_depth if self.max_depth is not None else -1,
        )
        self.context_info_tree = ContextTree(self.jab_wrapper, max_depth=self.max_depth)

    @keyword
    def application_refresh(self):
        """Refresh application element tree

        Might be required action after application element
        structure changes after window refresh.
        """
        self._application_refresh(force=True)

    @keyword
    def press_keys(self, *keys):
        """Press multiple keys down simultaneously

        See `Desktop`_ library documentation for supported keys

        .. _Desktop: https://rpaframework.org/libraries/desktop/index.html

        :param keys: keys to press
        """
        DESKTOP.press_keys(*keys)

    @keyword
    def print_element_tree(self, filename: Optional[str] = None):
        """Print current element into log and possibly into a file

        :param filename: filepath to save element tree
        :return: element tree
        """
        tree = repr(self.context_info_tree)
        self.logger.info(tree)
        if filename:
            with open(filename, "w", encoding="utf-8") as f:
                f.write(tree)
            self.logger.info("Context tree written to file '%s'", filename)
        return tree

    @keyword
    def get_locator_tree(self):
        """Return Java locator tree as list of objects.

        Mostly relevant object properties are:

            - ancestry
            - role
            - name
            - description
            - indexInParent

        :return: list of objects
        """
        return self.context_info_tree.get_search_element_tree()

    @keyword
    def print_locator_tree(self, filename: Optional[str] = None):
        """Print current Java window locator list into log and possibly
        into a file.

        :param filename: filepath to save locator tree
        :return: locator tree
        """
        search_tree = self.get_locator_tree()
        tree = "\n".join(
            [
                f"{'| ' * item.ancestry}role:{item.role} and name:{item.name} and "
                f"description:{item.description} "
                f"and indexInParent:{item.indexInParent}"
                for item in search_tree
            ]
        )
        self.logger.info(tree)
        if filename:
            with open(filename, "w", encoding="utf-8") as f:
                f.write(tree)
            self.logger.info("Locator tree written to file '%s'", filename)
        return tree

    @keyword
    def select_menu(self, menu: str, menuitem: str):
        """Select menu by clicking menu elements

        :param menu: name of the menu
        :param menuitem: name of the menu item
        """
        self.click_element(f"role:menu and name:{menu}")
        self.click_element(f"role:menu item and name:{menuitem}")

    @keyword
    def click_push_button(self, button_name: str):
        """Click element of role `push button`

        :param button_name: name of the button to click
        """
        locator = f"role:push button and name:{button_name}"
        self.click_element(locator)

    @keyword
    def shutdown_jab(self):
        """Call Java Access Bridge process shutdown"""
        self.jab_wrapper.shutdown()

    @keyword
    def get_version_info(self):
        """Get Java Access Bridge version information"""
        version_info = self.jab_wrapper.get_version_info()
        self.logger.info(
            "VMversion=%s; BridgeJavaClassVersion=%s; BridgeJavaDLLVersion=%s; BridgeWinDLLVersion=%s",  # noqa: E501
            version_info.VMversion,
            version_info.bridgeJavaClassVersion,
            version_info.bridgeJavaDLLVersion,
            version_info.bridgeWinDLLVersion,
        )
        return version_info

    @keyword
    def read_table(self, locator: LocatorType, visible_only: bool = True):
        """Return Java table as list of lists (rows containing columns).

        Each cell element is represented by ``JavaElement`` class.

        :param locator: locator to match element with type of table
        :param visible_only: return all the children when this is `False`
        :return: list of lists

        Example.

        .. code:: python

            table = java.read_table(locator_table)
            for row in table:
                for cell in row:
                    if cell.role == "check box":
                        print(cell.row, cell.col, str(cell.checked))
                    else:
                        print(cell.row, cell.col, cell.name)
        """
        # Refresh the table before encapsulating it in a `JavaElement` object.
        table = self._get_matching_element(
            locator, as_java_element=True, refresh=self.ignore_callbacks
        )
        if visible_only:
            active_children = table.visible_children
        else:
            active_children = table.node.children
            # Reset the column counter when dealing with all the children this time.
            # NOTE(cmin764; 4 Oct 23): Normally, we would set `table.column_count` to
            #  `None` first, but node's all children are not having any visible
            #  coordinates (-1), thus not reliable when computing the number of columns
            #  again if we don't keep the previously set counter.
            table.set_rows_columns(active_children)
        self.logger.info("Read table: %s", table)

        column_count = table.column_count
        if not column_count:
            raise InvalidLocatorError(
                f"Locator {locator!r} does not match the 'table' element"
            )

        table_elements = []
        for index, active_child in enumerate(active_children):
            java_elem = JavaElement(
                active_child,
                scaling_factor=self.display_scale_factor,
                index=index,
                column_count=column_count,
            )
            table_elements.append(java_elem)
        table_rows = [
            table_elements[idx : idx + column_count]
            for idx in range(0, len(table_elements), column_count)
        ]
        return table_rows

    @keyword
    def close_java_window(self):
        """Close active Java window which has been accessed
        via ```Select Window`` keyword.
        """
        if not self.pid:
            raise JavaWindowNotFound()
        os.system(f"taskkill /F /T /PID {self.pid}")

    @keyword
    def list_java_windows(self):
        """List all available Java windows.

        `JavaWindow` object contains following properties:

            - Java process id (pid)
            - Java window title
            - Java window handle (hwnd)

        The pid and title can be used to get control of the
        Java process by.

        :return: list of `JavaWindow` objects

        **Python example.**

        .. code:: python

            window_list = java.list_java_windows()
            # By looping window list
            for window in window_list:
                if window.title == "my java window title":
                    logging.info("Java window found")
                    java.select_window_by_pid(window.pid)
            # Directly accessing
            if len(window_list) == 1:
                java.select_window_by_pid(window_list[0].pid)

        **Robot Framework example.**

        .. code:: robotframework

            @{window_list}=    List Java Windows
            FOR    ${window}    IN    @{window_list}
                IF    "${window.title}" == "my java window title"
                    Select Window By PID    ${window.pid}
                END
            END
            IF    len($window_list)==1
                Select Window By PID    ${window_list[0].pid}
            END
        """
        if self.jab_wrapper is None:
            self._initialize()
        window_list = self.jab_wrapper.get_windows()
        return window_list

    @keyword
    def select_window_by_title(
        self, title: str, bring_foreground: bool = True, timeout: int = 30
    ):
        """Selects Java application window as target for the automation using
        Java window title.

        :param title: application window title
        :param bring_foreground: if application is brought to foreground or not
        :param timeout: selection timeout
        """
        self._select_window(
            title=title, pid=None, bring_foreground=bring_foreground, timeout=timeout
        )

    @keyword
    def select_window_by_pid(
        self, pid: int, bring_foreground: bool = True, timeout: int = 30
    ):
        """Selects Java application window as target for the automation using
        Java process ID (pid).

        :param pid: application process id
        :param bring_foreground: if application is brought to foreground or not
        :param timeout: selection timeout
        """
        self._select_window(pid=pid, bring_foreground=bring_foreground, timeout=timeout)

    def _find_window(self, title: Optional[str] = None, pid: Optional[int] = None):
        window_found = False
        try:
            if title:
                self.pid = self.jab_wrapper.switch_window_by_title(title)
            elif pid:
                self.pid = self.jab_wrapper.switch_window_by_pid(pid)
            window_found = True
        except Exception:  # pylint: disable=broad-except
            pass
        return window_found

    def _select_window(
        self,
        title: Optional[str] = None,
        pid: Optional[int] = None,
        bring_foreground: bool = True,
        timeout: int = 30,
    ):
        if self.jab_wrapper is None:
            self._initialize()

        window_found = False
        interval = float(0.5)
        end_time = time.time() + float(timeout)
        while time.time() <= end_time:
            start = time.time()
            window_found = self._find_window(title, pid)
            if window_found:
                break
            duration = time.time() - start
            if duration < interval:
                time.sleep(interval - duration)

        if not window_found:
            if title:
                raise ValueError("Did not find window with title '%s'" % title)
            elif pid:
                raise ValueError("Did not find window with pid '%s'" % pid)
            else:
                raise ValueError(
                    "Window selection was called without 'title' or 'pid'."
                    " Can't select window."
                )

        if not self.version_printed:
            self.get_version_info()
            self.version_printed = True

        if bring_foreground:
            self._bring_window_to_foreground()

        self._application_refresh(force=True)

        return self.pid

    def _bring_window_to_foreground(self):
        # correct window identifiable by the PID
        if self.pid:
            handle = self.jab_wrapper.get_current_windows_handle()
            # pylint: disable=c-extension-no-member
            win32gui.ShowWindow(handle, win32con.SW_SHOW)
            # pylint: disable=c-extension-no-member
            win32gui.SetForegroundWindow(handle)
        else:
            self.logger.warning(
                "Window pid is unknown, can't bring to window to foreground."
            )

    @keyword
    def wait_until_element_exists(self, locator: str, timeout: int = 10):
        """Wait until element(s) matching the locator are found within
        given timeout or raises ``ElementNotFound`` exception.

        :param locator: locator to match element
        :param timeout: timeout in seconds to find element
        :return: element(s) if found
        """
        elements = None
        end_time = time.time() + float(timeout)
        while time.time() <= end_time:
            start_time = time.time()
            elements = self._find_elements(locator)
            if len(elements) > 0:
                break
            self._application_refresh()
            # sleep at least a second per iteration
            duration = time.time() - start_time
            if duration < 1.0:
                time.sleep(1.0 - duration)
        if elements is None or len(elements) == 0:
            # print element tree when result is not expected
            self.print_element_tree()
            raise ElementNotFound("Locator '%s' did not match any elements" % locator)
        return elements

    @keyword
    def refresh_element(self, locator: LocatorType, index: int = 0) -> JavaElement:
        """Refresh an element alone.

        This will ensure the latest data is available in the targeted element, thus
        gaining speed when dealing with big apps that won't require an entire global
        refresh. The obtained Java element is returned.

        :param locator: element to refresh
        :param index: target element index if multiple are returned
        :returns: the Java element found by the passed locator
        """
        elem = self._get_matching_element(locator, index, as_java_element=True)
        self.logger.info("Refreshing element: %r", elem)
        elem.node.refresh()
        return elem



from datetime import date as datetime_date
import logging
from typing import Union, List, Dict, Optional

import pendulum as pdl
from pendulum.parsing.exceptions import ParserError
from pendulum.datetime import DateTime as PendulumDateTime

from robot.api.deco import keyword, library
import holidays

parsing_error_message = """Could not parse date '%s'.

You can use `Create Time` keyword to construct valid
calendar object by giving date and time as string in corresponding
date format. See https://pendulum.eustace.io/docs/#tokens for
valid tokens for the date format.
"""

DTFormat = Union[str, datetime_date, PendulumDateTime]


@library(scope="GLOBAL", doc_format="REST")
class Calendar:
    """Library for handling different operations for date and time
    handling especially in business days and holiday contexts.

    Utilizing  `pendulum <https://pypi.org/project/pendulum/>`_ and
    `holidays <https://pypi.org/project/holidays/>`_ packages.

    Library is by default using days from Monday to Friday as business
    days, but that can be changed by giving list of weekdays to
    ``Set Business Days`` keyword. A weekday is given as a integer, the
    0 for Sunday and 6 for Saturday.

    Common country holidays are respected when getting next and previous
    business days, but custom holidays can be added into consideration
    using keyword ``Add Custom Holidays`` keyword.

    Some dates containing for example month names are in English (en), but
    the locale of the library can be changed with keyword ``Set Locale`` or
    for specific keyword if that has a ``locale`` parameter.
    """

    def __init__(self) -> None:
        self.logger = logging.getLogger(__name__)
        self.DEFAULT_BUSINESS_DAYS = [0, 1, 2, 3, 4]  # Monday - Friday as a constant
        self.BUSINESS_DAYS = self.DEFAULT_BUSINESS_DAYS.copy()  # Working copy
        self.custom_holidays = holidays.HolidayBase()

    @keyword
    def reset_business_days(self) -> None:
        """Reset business days back to default Monday-Friday schedule."""
        self.BUSINESS_DAYS = self.DEFAULT_BUSINESS_DAYS.copy()

    @keyword
    def set_locale(self, locale_name: str) -> str:
        """Set locale globally for the library

        :param locale_name: name of the locale
        :return: name of the previous locale

        Python example.

        .. code-block:: python

            library = Calendar()
            library.set_locale("es")
            now = library.time_now(return_format="dddd DD MMMM YYYY")
            # now == "jueves 09 marzo 2023"
            library.set_locale("en")
            now = library.time_now(return_format="dddd DD MMMM YYYY")
            # now == "Thursday 09 March 2023"

        Robot Framework example.

        .. code-block:: robotframework

            Set Locale   es
            ${now}=  Time Now  return_format=dddd DD MMMM YYYY
            # ${now} == "jueves 09 marzo 2023"
            Set Locale   en
            ${now}=  Time Now  return_format=dddd DD MMMM YYYY
            # ${now} == "Thursday 09 March 2023"

        """
        previous = pdl.get_locale()
        pdl.set_locale(locale_name)
        return previous

    @keyword
    def reset_custom_holidays(self) -> None:
        """Reset custom holiday list into empty list."""
        self.custom_holidays = holidays.HolidayBase()

    @keyword
    def add_custom_holidays(self, days: Union[DTFormat, List[DTFormat]]) -> List:
        """Add a day or list of days which are considered as holidays
        in addition to country specific holidays when calculating

        :param days: string or list of dates to consider as holidays
        :return: list of current custom holidays

        Python example.

        .. code-block:: python

            library = Calendar()
            custom_holidays = library.add_custom_holidays("2023-03-08")
            # custom_holidays == ["2023-03-08"]
            custom_holidays = library.add_custom_holidays([
                "2023-03-09", "2023-03-10"
            ])
            # custom_holidays == ["2023-03-08", "2023-03-09", "2023-03-10"]

        Robot Framework example.

        .. code-block:: robotframework

            @{custom_holidays}=   Add Custom Holidays   2023-03-08
            # ${custom_holidays} == ["2023-03-08"]
            @{more_holidays}=   Create List   2023-03-09   2023-03-10
            @{custom_holidays}=   Add Custom Holidays   ${more_holidays}
            # ${custom_holidays} == ["2023-03-08", "2023-03-09", "2023-03-10"]

        """
        if isinstance(days, str):
            self.custom_holidays.append(days.split(","))
        else:
            self.custom_holidays.append(days)
        return self.custom_holidays

    @keyword
    def set_business_days(self, days: List[int]) -> List:
        """Set weekdays which are considered as business days
        for calculating previous and next business day.

        :param days: list of integers denoting weekdays
        :return: previous list of weekdays

        Python example.

        .. code-block:: python

            # set 4 day work week
            previous = Calendar().set_business_days([1,2,3,4])
            # previous == [1,2,3,4,5]

        Robot Framework example.

        .. code-block:: robotframework

            @{4days}=   Create List   1  2  3  4
            @{previous}=    Set Business Days  ${days}
            # ${previous} == [1,2,3,4,5]

        """
        previous = self.BUSINESS_DAYS
        self.BUSINESS_DAYS = days
        return previous

    @keyword
    def time_difference(
        self,
        start_date: Union[DTFormat, str],
        end_date: Union[DTFormat, str],
        start_timezone: Optional[str] = None,
        end_timezone: Optional[str] = None,
    ) -> Dict:
        """Compare 2 dates and get the time difference.

        Returned dictionary contains following properties:

            - end_date_is_later, `True` if end_date is more recent
              than start_date, otherwise `False`
            - years, time difference in years
            - months, time difference in months
            - days, time difference in days
            - hours, time difference in hours (in addition to the days)
            - minutes, time difference in minutes (in addition to the hours)
            - seconds, time difference in seconds (in addition to the minutes)

        :param start_date: starting date for the comparison
        :param end_date: ending date for the comparison
        :param start_timezone: timezone for the starting date, defaults to None
        :param end_timezone: timezone for the ending date, defaults to None
        :return: dictionary containing comparison result

        Python example.

        .. code-block:: python

            diff = Calendar().time_difference(
                "1975-05-22T18:00:00",
                "1975-05-22T22:45:30"
            )
            # diff['end_date_is_later'] == True
            # diff['days'] == 0
            # diff['hours'] == 4
            # diff['minutes'] == 45
            # diff['seconds'] == 30

        Robot Framework example.

        .. code-block:: robotframework

            &{diff}=    Time Difference  1975-05-22T18:00:00  1975-05-22T22:45:30
            # ${diff}[end_date_is_later] == True
            # ${diff}[days] == 0
            # ${diff}[hours] == 4
            # ${diff}[minutes] == 45
            # ${diff}[seconds] == 30

        """

        if isinstance(start_date, str):
            start_d = self._parse_datetime_string_to_pendulum_datetime(
                start_date, timezone=start_timezone
            )
        else:
            start_d = start_date
        if isinstance(end_date, str):
            end_d = self._parse_datetime_string_to_pendulum_datetime(
                end_date, timezone=end_timezone
            )
        else:
            end_d = end_date

        diff = end_d - start_d
        modifier_for_seconds = 1 if diff.seconds >= 0 else -1
        return {
            "end_date_is_later": end_d > start_d,
            "years": diff.years,
            "months": diff.months,
            "days": diff.days,
            "hours": diff.hours,
            "minutes": diff.minutes,
            "seconds": (
                diff.seconds
                if abs(diff.seconds) <= 60
                else abs(diff.seconds) % 60 * modifier_for_seconds
            ),
        }

    @keyword
    def create_time(
        self,
        date_string: str,
        date_format_in: Optional[str] = None,
        timezone: Optional[str] = None,
        date_format_out: Optional[str] = None,
    ) -> Union[PendulumDateTime, str]:
        """This keyword tries to construct valid calendar
        instance from given date string and its expected date
        format.

        See https://pendulum.eustace.io/docs/#tokens for
        valid tokens for the date format. Tokens are
        used to form correct date and time format.

        :param date_string: for example. "22 May 19"
        :param date_format_in: for example. "DD MMM YY"
        :param timezone: default timezone is "UTC"
        :param date_format_out: for example. "DD-MM-YY"
        :return: set datetime as an object or string
         if `date_format_out` has been set

        Python example.

        .. code-block:: python

            date = Calendar().create_time(
                "22 May 19",
                "DD MMM YY"
            )

        Robot Framework example.

        .. code-block:: robotframework

            ${date}=  Create Time
            ...  22 May 19
            ...  DD MMM YY

        """
        result = None
        if date_format_in:
            result = pdl.from_format(date_string, date_format_in, tz=timezone)
        else:
            result = self._parse_datetime_string_to_pendulum_datetime(
                date_string, timezone=timezone
            )
        return result.format(date_format_out) if date_format_out else result

    @keyword
    def time_now(
        self,
        timezone: Optional[str] = None,
        return_format: str = "YYYY-MM-DD",
    ) -> PendulumDateTime:
        """Return current date and time

        :param timezone: optional, for example. "America/Boston"
        :param return_format: dates can be formatted for the resulting
         list, defaults to "YYYY-MM-DD"
        :return: current datetime as an object

        Python example.

        .. code-block:: python

            now = Calendar().time_now("Europe/Helsinki")

        Robot Framework example.

        .. code-block:: robotframework

            ${now}=  Time Now   Europe/Helsinki
        """
        now = pdl.now(tz=timezone)
        if return_format:
            now = now.format(return_format)
        return now

    def _parse_datetime_string_to_pendulum_datetime(
        self, date_string: DTFormat, timezone: str = None
    ):
        arguments = {"text": date_string, "strict": False}
        if timezone:
            arguments["tz"] = timezone
        try:
            result = pdl.parse(**arguments)
            return result
        except ParserError as err:
            raise ValueError(parsing_error_message % date_string) from err

    @keyword
    def time_difference_in_months(
        self,
        start_date: DTFormat,
        end_date: DTFormat,
        start_timezone: Optional[str] = None,
        end_timezone: Optional[str] = None,
    ):
        """Return time difference of dates in months.

        :param start_date: the start date
        :param end_date: the end date
        :param start_timezone: timezone for the start date,
         defaults to None
        :param end_timezone: timezone for the end date,
         defaults to None
        :return: difference in months

        Python example.

        .. code-block:: python

            diff = Calendar().time_difference_in_months(
                "2022-05-21T22:00:00",
                "2023-08-21T22:00:00"
            )
            # diff == 15

        Robot Framework example.

        .. code-block:: robotframework

            ${diff}=  Time Difference In Months
            ...  2022-05-21T22:00:00
            ...  2023-08-21T22:00:00
            # ${diff} == 15

        """
        start_date_dt = self.create_time(start_date, timezone=start_timezone)
        end_date_dt = self.create_time(end_date, timezone=end_timezone)
        return end_date_dt.diff(start_date_dt).in_months()

    @keyword
    def time_difference_in_days(
        self,
        start_date: DTFormat,
        end_date: DTFormat,
        start_timezone: Optional[str] = None,
        end_timezone: Optional[str] = None,
    ):
        """Return the time difference of dates in days.

        :param start_date: the start date
        :param end_date: the end date
        :param start_timezone: timezone for the start date,
         defaults to None
        :param end_timezone: timezone for the end date,
         defaults to None
        :return: difference in days

        Python example.

        .. code-block:: python

            diff = Calendar().time_difference_in_days(
                "2023-05-21",
                "2023-05-29"
            )
            # diff == 8

        Robot Framework example.

        .. code-block:: robotframework

            ${diff}=  Time Difference In Days
            ...  2023-05-21
            ...  2023-05-29
            # ${diff} == 8

        """
        start_date_dt = self.create_time(start_date, timezone=start_timezone)
        end_date_dt = self.create_time(end_date, timezone=end_timezone)
        return end_date_dt.diff(start_date_dt).in_days()

    @keyword
    def time_difference_in_hours(
        self,
        start_date: DTFormat,
        end_date: DTFormat,
        start_timezone: Optional[str] = None,
        end_timezone: Optional[str] = None,
    ):
        """Return the time difference of dates in hours.

        :param start_date: the start date
        :param end_date: the end date
        :param start_timezone: timezone for the start date,
         defaults to None
        :param end_timezone: timezone for the end date,
         defaults to None
        :return: difference in hours

        Python example.

        .. code-block:: python

            diff = Calendar().time_difference_in_hours(
                "2023-08-21T22:00:00",
                "2023-08-22T04:00:00"
            )
            # diff == 6

        Robot Framework example.

        .. code-block:: robotframework

            ${diff}=  Time Difference In Hours
            ...  2023-08-21T22:00:00
            ...  2023-08-22T04:00:00
            # ${diff} == 6

        """
        start_date_dt = self.create_time(start_date, timezone=start_timezone)
        end_date_dt = self.create_time(end_date, timezone=end_timezone)
        return end_date_dt.diff(start_date_dt).in_hours()

    @keyword
    def time_difference_in_minutes(
        self,
        start_date: DTFormat,
        end_date: DTFormat,
        start_timezone: Optional[str] = None,
        end_timezone: Optional[str] = None,
    ):
        """Return the time difference of dates in minutes.

        :param start_date: the start date
        :param end_date: the end date
        :param start_timezone: timezone for the start date,
         defaults to None
        :param end_timezone: timezone for the end date,
         defaults to None
        :return: difference in minutes

        Python example.

        .. code-block:: python

            diff = Calendar().time_difference_in_minutes(
                "12:30",
                "16:35"
            )
            # diff == 245

        Robot Framework example.

        .. code-block:: robotframework

            ${diff}=  Time Difference In Minutes
            ...  12:30
            ...  16:35
            # ${diff} == 245
        """
        start_date_dt = self.create_time(start_date, timezone=start_timezone)
        end_date_dt = self.create_time(end_date, timezone=end_timezone)
        return end_date_dt.diff(start_date_dt).in_minutes()

    @keyword
    def time_difference_between_timezones(
        self,
        start_timezone: str,
        end_timezone: str,
    ):
        """Return the hour difference between timezones.

        :param start_timezone: first timezone
        :param end_timezone: second timezone
        :return: hour difference between the timezones

        Python example.

        .. code-block:: python

            diff = Calendar().time_difference_between_timezones(
                "America/New_York",
                "Europe/Helsinki"
            )
            # diff == 7

        Robot Framework example.

        .. code-block:: robotframework

            ${diff}=  Time Difference Between Timezones
            ...  America/New_York
            ...  Europe/Helsinki
            # ${diff} == 7

        """
        start_date_dt = pdl.datetime(2023, 1, 1, tz=start_timezone)
        end_date_dt = pdl.datetime(2023, 1, 1, tz=end_timezone)
        return end_date_dt.diff(start_date_dt).in_hours()

    @keyword
    def return_previous_business_day(
        self,
        date: DTFormat,
        country: Optional[str] = None,
        return_format: str = "YYYY-MM-DD",
        locale: Optional[str] = None,
    ):
        """Return the previous business day.

        :param date: day of origin
        :param country: country code, default `None`
        :param return_format: dates can be formatted for the resulting
         list, defaults to "YYYY-MM-DD"
        :param locale: name of the locale
        :return: the previous business day from day of origin

        Python example.

        .. code-block:: python

            prev_business = Calendar().return_previous_business_day("2023-01-09", "FI")
            # prev == "2023-01-05"

        Robot Framework example.

        .. code-block:: robotframework

            ${previous_business}=  Return Previous Business Day  2023-01-09  FI
            # ${previous_business} == "2023-01-05"
        """
        return self._return_business_day(date, country, return_format, locale, -1)

    @keyword
    def return_next_business_day(
        self,
        date: DTFormat,
        country: Optional[str] = None,
        return_format: str = "YYYY-MM-DD",
        locale: Optional[str] = None,
    ):
        """Return the next business day.

        :param date: day of origin
        :param country: country code, default `None`
        :param return_format: dates can be formatted for the resulting
         list, defaults to "YYYY-MM-DD"
        :param locale: name of the locale
        :return: the next business day from day of origin

        Python example.

        .. code-block:: python

            next_business = Calendar().return_next_business_day("2023-01-05", "FI")
            # next_business == "2023-01-09"

        Robot Framework example.

        .. code-block:: robotframework

            ${next_business}=  Return Next Business Day  2023-01-05  FI
            # ${next_business} == "2023-01-09"
        """
        return self._return_business_day(date, country, return_format, locale, 1)

    def _return_business_day(
        self,
        given_date: DTFormat,
        country: Optional[str] = None,
        return_format: Optional[str] = None,
        locale: Optional[str] = None,
        direction: int = -1,
    ):
        if isinstance(given_date, str):
            given_dt = pdl.parse(given_date, strict=False)
        else:
            given_dt = given_date

        current_dt = given_dt
        years = [given_dt.year - 1, given_dt.year, given_dt.year + 1]
        holiday_list = self.return_holidays(years, country)

        # Skip the current date for previous/next business day
        current_dt = current_dt.add(days=direction)

        while True:
            current_date = pdl.date(current_dt.year, current_dt.month, current_dt.day)
            weekday = current_dt.day_of_week

            self.logger.debug(f"Checking date: {current_date}, weekday: {weekday}")
            self.logger.debug(f"Business days: {self.BUSINESS_DAYS}")

            # Check if it's a business day
            is_business_day = weekday in self.BUSINESS_DAYS

            self.logger.debug(f"Is business day? {is_business_day}")

            # Check if it's not a holiday (if country is specified)
            if country and is_business_day:
                is_holiday = current_date in holiday_list
                is_business_day = not is_holiday
                self.logger.debug(f"Is holiday? {is_holiday}")

            if is_business_day:
                break

            current_dt = current_dt.add(days=direction)

        if return_format:
            return current_dt.format(fmt=return_format, locale=locale)
        return current_dt

    @keyword
    def return_holidays(
        self, years: Union[int, List[int]], country: Optional[str] = None
    ) -> Dict:
        """Return holidays for a country. If country is not given
        then only custom holidays are returned.

        :param years: single year or list of years to list holidays for
        :param country: country code, default `None`
        :return: holidays in a dictionary, the key is the date and the
         value is name of the holiday

        Python example.

        .. code-block:: python

            holidays = Calendar().return_holidays(2023, "FI")
            for date, holiday_name in holidays.items():
                print(f"{date} is {holiday_name}")

        Robot Framework example.

        .. code-block:: robotframework

            &{holidays}=  Return Holidays  2023  FI
            FOR  ${date}  IN   @{holidays.keys()}
                Log To Console   ${date} is ${holidays}[${date}]
            END

        """
        if country:
            holiday_list = holidays.country_holidays(country=country, years=years)
        else:
            holiday_list = holidays.HolidayBase(years=years)
        # add custom holidays
        holiday_list += self.custom_holidays

        return holiday_list

    @keyword
    def first_business_day_of_the_month(
        self, date: DTFormat, country: Optional[str] = None
    ):
        """Return first business day of the month.

        If `country` is not given then holidays are not considered.

        :param date: date describing the month
        :param country: country code, default `None`
        :return: first business of the month

        Python example.

        .. code-block:: python

            first_day = Calendar().first_business_day_of_the_month("2024-06-01")
            # first_day == "2024-06-03"

        Robot Framework example.

        .. code-block:: robotframework

            ${first_day}=  First Business Day of the Month  2024-06-01
            # ${first_day} == "2024-06-03"
        """
        result = None
        if isinstance(date, str):
            given_dt = pdl.parse(date, strict=False)
        else:
            given_dt = date
        year, current_month = given_dt.year, given_dt.month
        day = 2
        for _ in range(32):
            day_to_check = pdl.date(year, current_month, day)
            result = self.return_previous_business_day(
                day_to_check, country=country, return_format=None
            )
            if result.month == current_month:
                break
            else:
                day += 1
        return result

    @keyword
    def last_business_day_of_the_month(
        self, date: DTFormat, country: Optional[str] = None
    ):
        """Return last business day of the month.

        If `country` is not given then holidays are not considered.

        :param date: date describing the month
        :param country: country code, default `None`
        :return: last business day of the month

        Python example.

        .. code-block:: python

            last_day = Calendar().last_business_day_of_the_month("2023-12-01")
            # last_day == "2023-12-29"

        Robot Framework example.

        .. code-block:: robotframework

            ${last_day}=  Last Business Day of the Month  2023-12-01
            # ${last_day} == "2023-12-29"

        """
        result = None
        if isinstance(date, str):
            given_dt = pdl.parse(date, strict=False)
        else:
            given_dt = date
        year, current_month = given_dt.year, given_dt.month
        day = 1
        for _ in range(32):
            try:
                if day == 1:
                    next_month = given_dt.set(day=1).add(months=1)
                    day_to_check = pdl.date(next_month.year, next_month.month, day)
                    day = 31
                else:
                    day_to_check = pdl.date(year, current_month, day)
                result = self.return_previous_business_day(
                    day_to_check, country=country, return_format=None
                )
                if result.month == current_month:
                    break
                else:
                    day -= 1
            except ValueError:
                day -= 1
        return result

    @keyword
    def sort_list_of_dates(
        self,
        dates: List[DTFormat],
        return_format: Optional[str] = None,
        reverse: bool = False,
    ) -> List:
        """Sort list of dates.

        :param dates: list of dates to sort
        :param return_format: dates can be formatted for the resulting
         list
        :param reverse: `True` return latest to oldest, defaults to `False`,
         which means order from oldest to latest
        :return: list of sorted dates

        Python example.

        .. code-block:: python

            datelist = [
                "2023-07-02 12:02:31",
                "2023-07-03 12:02:35",
                "2023-07-03 12:02:31"
            ]
            sorted = Calendar().sort_list_of_dates(datelist)
            # sorted[0] == "2023-07-03 12:02:35"
            # sorted[-1] == "2023-07-02 12:02:31"
            sorted = Calendar().sort_list_of_dates(datelist, reverse=True)
            # sorted[0] == "2023-07-02 12:02:31"
            # sorted[-1] == "2023-07-03 12:02:35"

        Robot Framework example.

        .. code-block:: robotframework

            @{datelist}=  Create List
            ...   2023-07-02 12:02:31
            ...   2023-07-03 12:02:35
            ...   2023-07-03 12:02:31
            ${sorted}=  Sort List Of Dates   ${datelist}
            # ${sorted}[0] == "2023-07-03 12:02:35"
            # ${sorted}[-1] == "2023-07-02 12:02:31"
            ${sorted}=  Sort List Of Dates   ${datelist}  reverse=True
            # ${sorted}[0] == "2023-07-02 12:02:31"
            # ${sorted}[-1] == "2023-07-03 12:02:35"

        """
        dt_list = [self.create_time(date) for date in dates]
        result = sorted(dt_list, reverse=reverse)
        return [d.format(return_format) for d in result] if return_format else result

    @keyword(name="Compare Times ${time1} < ${time2}")
    def _rfw_compare_time_before(self, time1: DTFormat, time2: DTFormat):
        """Compares given times and returns `True` if `time2`
        is more recent than `time1`.

        :param time1: first time for comparison
        :param time2: second time for comparison
        :return: `True` if `time2` is more recent than `time1`

        Robot Framework example.

        .. code-block:: robotframework

            ${recent}=  Compare Times 2023-03-09 15:50 < 2023-03-09 15:59
            IF  ${recent}
                Log  2023-03-09 15:59 is more recent
            END

        """
        return self.compare_times(time1, time2)

    @keyword(name="Compare Times ${time1} > ${time2}")
    def _rfw_compare_time_after(self, time1: DTFormat, time2: DTFormat):
        """Compares given times and returns `True` if `time1`
        is more recent than `time2`.

        :param time1: first time for comparison
        :param time2: second time for comparison
        :return: `True` if `time1` is more recent than `time2`

        Robot Framework example.

        .. code-block:: robotframework

            ${recent}=  Compare Times 2023-03-09 15:59 > 2023-03-09 15:58
            IF  ${recent}
                Log  2023-03-09 15:59 is more recent
            END

        """
        return not self.compare_times(time1, time2)

    @keyword
    def compare_times(self, time1: DTFormat, time2: DTFormat):
        """Compares given times and returns `True` if `time2`
        is more recent than `time1`.

        :param time1: first time for comparison
        :param time2: second time for comparison
        :return: `True` if `time2` is more recent than `time1`

        Python example.

        .. code-block:: python

            recent = Calendar().compare_times("2023-03-09 13:02", "2023-03-09 13:47")
            if recent:
                print("2023-03-09 13:47 is more recent")

        Robot Framework example.

        .. code-block:: robotframework

            ${recent}=  Compare Times   2023-03-09 13:02   2023-03-09 13:47
            IF  ${recent}
                Log  2023-03-09 13:47 is more recent
            END

        """
        diff = self.time_difference(time1, time2)
        return diff["end_date_is_later"]

    @keyword
    def get_iso_calendar(self, date: DTFormat):
        """Get ISO calendar information for the given date.

        :parameter date: input date
        :return: ISO calendar object containing year, week number and weekday.

        Python example.

        .. code-block:: python

            iso_cal = Calendar().get_iso_calendar("2023-03-09")
            print(iso_cal.year)
            print(iso_cal.week)
            print(iso_cal.weekday)

        Robot Framework example.

        .. code-block:: robotframework

            ${iso_cal}=  Get ISO Calendar  2023-03-09
            ${iso_year}=  Set Variable  ${iso_cal.year}
            ${iso_week}=  Set Variable  ${iso_cal.week}
            ${iso_weekday}=  Set Variable  ${iso_cal.weekday}
        """
        if isinstance(date, str):
            given_dt = pdl.parse(date, strict=False)
        else:
            given_dt = date
        return given_dt.isocalendar()

    @keyword
    def is_the_date_business_day(
        self, date: DTFormat, country: Optional[str] = None
    ) -> bool:
        """Is the date a business day in a country.

        If `country` is not given then holidays are not considered.

        :param date: input date
        :param country: country code, default `None`
        :return: `True` if the day is a business day, `False` if not

        Python example.

        .. code-block:: python

            for day in range(1,32):
                date = f"2023-1-{day}"
                is_business_day = Calendar().is_the_date_business_day(date, "FI")
                if is_business_day:
                    print(f'It is time for the work on {date}')
                else:
                    print(f'It is time to relax on {date}')

        Robot Framework example.

        .. code-block:: robotframework

            FOR  ${day}  IN RANGE  1  32
                ${date}=   Set Variable   2023-1-${day}
                ${is_business_day}=   Is the date business day  ${date}  FI
                IF   ${is_business_day}
                    Log To Console   It is time for the work on ${date}
                ELSE
                    Log To Console   It is time to relax on ${date}
                END
            END
        """
        if isinstance(date, str):
            given_dt = pdl.parse(date, strict=False)
        else:
            given_dt = date
        holiday_list = self.return_holidays(given_dt.year, country)
        is_business_day = False
        if given_dt.day_of_week in self.BUSINESS_DAYS:
            is_business_day = True
        if country and is_business_day:
            is_business_day = given_dt not in holiday_list
        return is_business_day

    @keyword(name="Is the ${date} Business Day in ${country}")
    def _rfw_is_the_day_business_day(self, date: DTFormat, country: str):
        """Is the date a business day in a country.

        :param date_in: input date
        :param country: country code
        :return: `True` if the day is a business day, `False` if not

        Robot Framework example.

        .. code-block:: robotframework

            ${is_business_day}=   Is the 2023-01-02 business day in FI
            IF   ${is_business_day}
                Log To Console   It is time for the work
            ELSE
                Log To Console   It is time to relax
            END
        """
        return self.is_the_date_business_day(date, country)

    @keyword
    def is_the_date_holiday(self, date: DTFormat, country: Optional[str] = None):
        """Is the date a holiday in a country.
        If `country` is not given then checks only if date is in custom holiday list.

        :param date_in: input date
        :param country: country code, default `None`
        :return: `True` if the day is a holiday, `False` if not

        Python example.

        .. code-block:: python

            is_holiday = Calendar().is_the_date_holiday("2022-12-26", "FI")
            if is_holiday:
                print('Time to relax')
            else:
                print('Time for the work')

        Robot Framework example.

        .. code-block:: robotframework

            ${is_holiday}=   Is the date holiday   2022-12-26   FI
            IF   ${is_holiday}
                Log  Time to relax
            ELSE
                Log  Time for the work
            END

        """
        if isinstance(date, str):
            given_dt = pdl.parse(date, strict=False)
        else:
            given_dt = date
        holiday_list = self.return_holidays(given_dt.year, country)
        return given_dt in holiday_list

    @keyword(name="Is the ${date} Holiday in ${country}")
    def _rfw_is_the_day_holiday(self, date: DTFormat, country: str):
        """Is the date a holiday in a country.

        :param date_in: input date
        :param country: country code
        :return: `True` if the day is a holiday, `False` if not

        Robot Framework example.

        .. code-block:: robotframework

            ${is_it}=   Is the 2022-12-26 holiday in FI
            IF   ${is_holiday}
                Log  Time to relax
            ELSE
                Log  Time for the work
            END

        """
        return self.is_the_date_holiday(date, country)



from datetime import datetime
from fnmatch import fnmatch
import logging
import os
import os.path
from pathlib import Path
import tarfile
from typing import Union, List
import zipfile


def convert_date(timestamp):
    if isinstance(timestamp, tuple):
        d = datetime(
            year=timestamp[0],
            month=timestamp[1],
            day=timestamp[2],
            hour=timestamp[3],
            minute=timestamp[4],
            second=timestamp[5],
        )
    else:
        d = datetime.utcfromtimestamp(timestamp)
    formatted_date = d.strftime("%d.%m.%Y %H:%M:%S")
    return formatted_date


def list_files_in_directory(folder, recursive=False, include=None, exclude=None):
    filelist = []
    for rootdir, _, files in os.walk(folder):
        for file in files:
            archive_absolute = os.path.join(rootdir, file)
            archive_relative = rootdir.replace(folder, "")
            archive_relative = os.path.join(archive_relative, file)
            if include and not fnmatch(archive_relative, include):
                continue
            if exclude and fnmatch(archive_relative, exclude):
                continue
            filelist.append((archive_absolute, archive_relative))
        if not recursive:
            break
    return filelist


class Archive:
    """`Archive` is a library for operating with ZIP and TAR packages.

    **Examples**

    .. code-block:: robotframework

       *** Settings ***
       Library  RPA.Archive

       *** Tasks ***
       Creating a ZIP archive
          Archive Folder With ZIP   ${CURDIR}${/}tasks  tasks.zip   recursive=True  include=*.robot  exclude=/.*
          @{files}                  List Archive             tasks.zip
          FOR  ${file}  IN  ${files}
             Log  ${file}
          END
          Add To Archive            .${/}..${/}missing.robot  tasks.zip
          &{info}                   Get Archive Info


    .. code-block:: python

        from RPA.Archive import Archive

        lib = Archive()
        lib.archive_folder_with_tar('./tasks', 'tasks.tar', recursive=True)
        files = lib.list_archive('tasks.tar')
        for file in files:
           print(file)
    """  # noqa: E501

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"

    def __init__(self):
        self.logger = logging.getLogger(__name__)

    def archive_folder_with_zip(
        self,
        folder: str,
        archive_name: str,
        recursive: bool = False,
        include: str = None,
        exclude: str = None,
        compression: str = "stored",
    ) -> None:
        # pylint: disable=C0301
        """Create a zip archive of a folder

        :param folder: name of the folder to archive
        :param archive_name: filename of the archive
        :param recursive: should sub directories be included, default is False
        :param include: define file pattern to include in the package, default is None which means all files are included
        :param exclude: define file pattern to exclude from the package, default is None
        :param compression: type of package compression method, default is "stored"
        :return: None

        This keyword creates an ZIP archive of a local folder. By default subdirectories are not
        included, but they can be included using the `recursive` argument.

        To include only certain files, like TXT files, the argument `include` can be used.
        Similarly to exclude certain files, like dotfiles, the argument `exclude` can be used.

        Compression methods:

        - stored, default
        - deflated
        - bzip2
        - lzma

        Example:

        .. code-block:: robotframework

            Archive Folder With Zip  ${CURDIR}${/}documents  mydocs.zip
            Archive Folder With Zip  ${CURDIR}${/}tasks      robottasks.zip   include=*.robot
            Archive Folder With Zip  ${CURDIR}${/}tasks      no_dotfiles.zip  exclude=/.*
            Archive Folder With Zip  ${CURDIR}${/}documents  documents.zip    recursive=True
            Archive Folder With Zip  ${CURDIR}               packagelzma.zip  compression=lzma
            Archive Folder With Zip  ${CURDIR}               bzipped.zip      compression=bzip2


        .. code-block:: python

            from RPA.Archive import Archive

            lib = Archive()
            lib.archive_folder_with_zip('./documents', 'mydocs.zip')
            lib.archive_folder_with_zip('./tasks', 'robottasks.zip', include='*.robot')
            lib.archive_folder_with_zip('./tasks', 'no_dotfiles.zip', exclude='/.*')
            lib.archive_folder_with_zip('./documents', 'documents.zip', recursive=True)
            lib.archive_folder_with_zip('./', 'packagelzma.zip', compression='lzma')
            lib.archive_folder_with_zip('./', 'bzipped.zip', compression='bzip2')
        """  # noqa: E501

        if compression == "stored":
            comp_method = zipfile.ZIP_STORED
        elif compression == "deflated":
            comp_method = zipfile.ZIP_DEFLATED
        elif compression == "bzip2":
            comp_method = zipfile.ZIP_BZIP2
        elif compression == "lzma":
            comp_method = zipfile.ZIP_LZMA
        else:
            raise ValueError("Unknown compression method")

        filelist = list_files_in_directory(folder, recursive, include, exclude)
        if len(filelist) == 0:
            raise ValueError("No files found to archive")

        with zipfile.ZipFile(
            file=archive_name,
            mode="w",
            compression=comp_method,
        ) as archive:
            for archive_absolute, archive_relative in filelist:
                archive.write(archive_absolute, arcname=archive_relative)

    def archive_folder_with_tar(
        self,
        folder: str,
        archive_name: str,
        recursive: bool = False,
        include: str = None,
        exclude: str = None,
    ) -> None:
        """Create a tar/tar.gz archive of a folder

        :param folder: name of the folder to archive
        :param archive_name: filename of the archive
        :param recursive: should sub directories be included, default is False
        :param include: define file pattern to include in the package, default is None which means all files are included
        :param exclude: define file pattern to exclude from the package, default is None
        :return: None

        This keyword creates an TAR or TAR.GZ archive of a local folder. Type of archive
        is determined by the file extension. By default subdirectories are not
        included, but they can included using `recursive` argument.

        To include only certain files, like TXT files, the argument `include` can be used.
        Similarly to exclude certain file, like dotfiles, the argument `exclude` can be used.

        Example:

        .. code-block:: robotframework

            Archive Folder With TAR  ${CURDIR}${/}documents  documents.tar
            Archive Folder With TAR  ${CURDIR}${/}tasks      tasks.tar.gz   include=*.robot
            Archive Folder With TAR  ${CURDIR}${/}tasks      tasks.tar      exclude=/.*
            Archive Folder With TAR  ${CURDIR}${/}documents  documents.tar  recursive=True
        """  # noqa: E501
        filelist = list_files_in_directory(folder, recursive, include, exclude)
        if len(filelist) == 0:
            raise ValueError("No files found to archive")

        with tarfile.TarFile(archive_name, "w") as archive:
            for archive_absolute, archive_relative in filelist:
                archive.add(archive_absolute, arcname=archive_relative)

    def add_to_archive(
        self,
        files: Union[List, str],
        archive_name: str,
        folder: str = None,
    ) -> None:
        """Add file(s) to the archive

        :param files: name of the file, or list of files, to add
        :param archive_name: filename of the archive
        :param folder: name of the folder where the file will be added, relative path within the archive
        :return: None

        This keyword adds a file or list of files into an existing archive. Files
        can be added to archive structure with relative path using argument `folder`.

        Example:

        .. code-block:: robotframework

            Add To Archive  extrafile.txt  myfiles.zip
            Add To Archive  stat.png       archive.tar.gz  images
            @{files}        Create List    filename1.txt   filename2.txt
            Add To Archive  ${files}       files.tar
        """  # noqa: E501
        files_to_add = []

        if isinstance(files, list):
            for filename in files:
                file_relative = Path(filename).name
                if folder:
                    file_relative = folder + os.path.sep + file_relative
                files_to_add.append((filename, file_relative))
        else:
            file_relative = Path(files).name
            if folder:
                file_relative = folder + os.path.sep + file_relative
            files_to_add.append((files, file_relative))

        if zipfile.is_zipfile(archive_name):
            with zipfile.ZipFile(archive_name, "a") as zip_archive:
                for filename, file_relative in files_to_add:
                    zip_archive.write(filename, arcname=file_relative)
        elif tarfile.is_tarfile(archive_name):
            with tarfile.TarFile(archive_name, "a") as tar_archive:
                for file_absolute, file_relative in files_to_add:
                    tar_archive.add(file_absolute, arcname=file_relative)

    def list_archive(self, archive_name: str) -> list:
        """List files in an archive

        :param archive_name: filename of the archive
        :return: dictionary variable containing the keys name, size, mtime, modified

        Returns list of files, where each file in a list is a dictionary
        with following attributes:

        - name
        - size
        - mtime
        - last modification time in format `%d.%m.%Y %H:%M:%S`

        Example:

        .. code-block:: robotframework

            @{files}   List Archive    myfiles.zip
            FOR  ${file}  IN   ${files}
                Log  ${file}[filename]
                Log  ${file}[size]
                Log  ${file}[mtime]
            END
        """
        filelist = []
        if zipfile.is_zipfile(archive_name):
            with zipfile.ZipFile(archive_name, "r") as f:
                members = f.infolist()
                for memb in members:
                    filelist.append(
                        {
                            "filename": memb.filename,
                            "size": memb.file_size,
                            "mtime": memb.date_time,
                            "modified": convert_date(memb.date_time),
                        }
                    )
        elif tarfile.is_tarfile(archive_name):
            with tarfile.TarFile(archive_name, "r") as f:
                members = f.getmembers()
                for memb in members:
                    filelist.append(
                        {
                            "name": memb.name,
                            "size": memb.size,
                            "mtime": memb.mtime,
                            "modified": convert_date(memb.mtime),
                        }
                    )
        return filelist

    def get_archive_info(self, archive_name: str) -> dict:
        """Get information about the archive

        :param archive_name: filename of the archive
        :return: dictionary variable containing the keys filename, filemode, size, mtime, modified

        Returns following file attributes in a dictionary:

        - filename
        - filemode
        - size
        - mtime
        - last modification time in format `%d.%m.%Y %H:%M:%S`

        Example:

        .. code-block:: robotframework

            &{archiveinfo}   Get Archive Info    myfiles.zip
        """  # noqa: E501
        archive_info = None
        st = os.stat(archive_name)
        if zipfile.is_zipfile(archive_name):
            with zipfile.ZipFile(archive_name, "r") as f:
                archive_info = {
                    "filename": f.filename,
                    "mode": f.mode,
                    "size": st.st_size,
                    "mtime": st.st_mtime,
                    "modified": convert_date(st.st_mtime),
                }
        elif tarfile.is_tarfile(archive_name):
            with tarfile.TarFile(archive_name, "r") as f:
                archive_info = {
                    "filename": f.name,
                    "mode": f.mode,
                    "size": st.st_size,
                    "mtime": st.st_mtime,
                    "modified": convert_date(st.st_mtime),
                }
        return archive_info

    def extract_archive(
        self, archive_name: str, path: str = None, members: Union[List, str] = None
    ) -> None:
        """Extract files from archive into local directory

        :param archive_name: filename of the archive
        :param path: filepath to extract file into, default is current working directory
        :param members: list of files to extract from archive, by default
            all files in archive are extracted
        :return: None

        This keyword supports extracting files from zip, tar and tar.gz archives.

        By default file is extracted into current working directory, but `path` argument
        can be set to define extraction path.

        Example:

        .. code-block:: robotframework

            Extract Archive    myfiles.zip   ${CURDIR}${/}extracted
            @{files}           Create List   filename1.txt    filename2.txt
            Extract Archive    archive.tar   C:${/}myfiles${/}  ${files}
        """  # noqa: E501
        root = Path(path) if path else Path.cwd()
        if members and not isinstance(members, list):
            members = [members]
        if zipfile.is_zipfile(archive_name):
            with zipfile.ZipFile(archive_name, "r") as f:
                if members:
                    f.extractall(path=root, members=members)
                else:
                    f.extractall(path=root)
        elif tarfile.is_tarfile(archive_name):
            members = map(tarfile.TarInfo, members) if members else None
            with tarfile.open(archive_name, "r") as f:
                if members:
                    f.extractall(path=root, members=members)
                else:
                    f.extractall(path=root)

    def extract_file_from_archive(
        self, filename: str, archive_name: str, path: str = None
    ) -> None:  # pylint: disable=C0301
        """Extract a file from archive into local directory

        :param filename: name of the file to extract
        :param archive_name: filename of the archive
        :param path: filepath to extract file into,
            default is current working directory
        :return: None

        This keyword supports extracting a file from zip, tar and tar.gz archives.

        By default file is extracted into current working directory,
        but `path` argument can be set to define extraction path.

        Example:

        .. code-block:: robotframework

            Extract File From Archive    extrafile.txt   myfiles.zip
            Extract File From Archive    background.png  images.tar.gz  ${CURDIR}${/}extracted
        """  # noqa: E501
        root = Path(path) if path else Path.cwd()
        if zipfile.is_zipfile(archive_name):
            with zipfile.ZipFile(archive_name, "r") as outfile:
                outfile.extract(filename, root)
        elif tarfile.is_tarfile(archive_name):
            with tarfile.open(archive_name, "r") as outfile:
                outfile.extract(filename, root)
